import os
import shutil
import unittest

import h5py
import nibabel as nib
import nibabel.testing as nib_testing
import numpy as np
import pydicom
import pydicom.data as pydd
import SimpleITK as sitk

from voxel.device import Device
from voxel.io.dicom import DicomReader
from voxel.io.format_io import ImageDataFormat
from voxel.io.nifti import NiftiReader, NiftiWriter
from voxel.med_volume import MedicalVolume
from voxel.utils import env

from . import util as ututils


class TestMedicalVolume(unittest.TestCase):
    _AFFINE = np.asarray(
        [
            [0.0, 0.0, 0.8, -171.41],
            [0.0, -0.3125, 0.0, 96.0154],
            [-0.3125, 0.0, 0.0, 47.0233],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )  # ('SI', 'AP', 'LR')

    _TEMP_PATH = os.path.join(ututils.TEMP_PATH, __name__)

    @classmethod
    def setUpClass(cls):
        os.makedirs(cls._TEMP_PATH, exist_ok=True)

    @classmethod
    def tearDownCls(cls):
        if os.path.isdir(cls._TEMP_PATH):
            shutil.rmtree(cls._TEMP_PATH)

    def test_reformat(self):
        mv = MedicalVolume(np.random.rand(10, 20, 30), self._AFFINE)
        new_orientation = tuple(x[::-1] for x in mv.orientation[::-1])

        mv2 = mv.reformat(new_orientation)
        assert mv2.orientation == new_orientation
        assert id(mv2) != id(mv)
        assert np.shares_memory(mv2._volume, mv._volume)

        mv2 = mv.reformat(new_orientation, inplace=True)
        assert mv2.orientation == new_orientation
        assert id(mv2) == id(mv)
        assert np.shares_memory(mv2._volume, mv._volume)

        mv2 = mv.reformat(mv.orientation)
        assert id(mv2) != id(mv)
        assert np.shares_memory(mv2._volume, mv._volume)

        mv2 = mv.reformat(mv.orientation, inplace=True)
        assert id(mv2) == id(mv)
        assert np.shares_memory(mv2._volume, mv._volume)

        mv2 = mv.reformat(new_orientation).reformat(mv.orientation)
        assert mv2.is_identical(mv)

        mv2 = mv.reformat(*new_orientation).reformat(*mv.orientation)
        assert mv2.is_identical(mv)

        with self.assertRaises(ValueError):
            mv.reformat(new_orientation, True)

    def test_reformat_as(self):
        mv = MedicalVolume(np.random.rand(10, 20, 30), self._AFFINE)
        mv2 = MedicalVolume(np.random.rand(10, 20, 30), self._AFFINE[:, (0, 2, 1, 3)])
        mv = mv.reformat_as(mv2)
        assert mv.orientation == mv2.orientation

    def test_reformat_header(self):
        volume = np.random.rand(10, 20, 30, 40)
        headers = ututils.build_dummy_headers(volume.shape[2:])
        mv = MedicalVolume(volume, self._AFFINE, headers=headers)
        new_orientation = tuple(x[::-1] for x in mv.orientation[::-1])

        mv2 = mv.reformat(new_orientation)
        assert mv2._headers.shape == (30, 1, 1, 40)

        mv2 = mv.clone()
        mv2.reformat(new_orientation, inplace=True)
        assert mv2._headers.shape == (30, 1, 1, 40)

        volume = np.random.rand(10, 20, 30, 40)
        headers = ututils.build_dummy_headers((volume.shape[2], 1))
        mv = MedicalVolume(volume, self._AFFINE, headers=headers)
        new_orientation = tuple(x[::-1] for x in mv.orientation[::-1])

        mv2 = mv.reformat(new_orientation)
        assert mv2._headers.shape == (30, 1, 1, 1)

    def test_metadata(self):
        field, field_val = "EchoTime", 4.0

        volume = np.random.rand(10, 20, 30, 40)
        headers = ututils.build_dummy_headers(volume.shape[2:], {field: field_val})
        mv_no_headers = MedicalVolume(volume, self._AFFINE)
        mv = MedicalVolume(volume, self._AFFINE, headers=headers)

        assert mv_no_headers.headers() is None
        assert mv_no_headers.headers(flatten=True) is None

        with self.assertRaises((KeyError, ValueError)):
            mv.get_metadata("foobar")
        assert mv.get_metadata("foobar", default=0) == 0

        echo_time = mv.get_metadata(field)
        assert echo_time == field_val

        new_val = 5.0
        mv2 = mv.clone(headers=True)
        mv2.set_metadata(field, new_val)
        assert mv.get_metadata(field, type(field_val)) == field_val
        assert mv2.get_metadata(field, type(new_val)) == new_val
        for h in mv2.headers(flatten=True):
            assert h[field].value == new_val

        new_val = 6.0
        mv2 = mv.clone(headers=True)
        mv2[..., 1].set_metadata(field, new_val)
        assert mv2[..., 0].get_metadata(field) == field_val
        assert mv2[..., 1].get_metadata(field) == new_val
        headers = mv2.headers()
        for h in headers[..., 0].flatten():
            assert h[field].value == field_val
        for h in headers[..., 1].flatten():
            assert h[field].value == new_val

        # Set metadata when volume has no headers.
        mv_nh = MedicalVolume(volume, self._AFFINE, headers=None)
        with self.assertRaises(ValueError):
            mv_nh.set_metadata("EchoTime", 40.0)
        with self.assertWarns(UserWarning):
            mv_nh.set_metadata("EchoTime", 40.0, force=True)
        assert mv_nh._headers.shape == (1,) * len(mv_nh.shape)
        assert mv_nh.get_metadata("EchoTime") == 40.0
        assert mv_nh[:1, :2, :3]._headers.shape == (1,) * len(mv_nh.shape)

    def test_apply_rescale(self):
        metadata = {"RescaleSlope": "2.5", "RescaleIntercept": "1.0"}
        volume = np.arange(9, dtype=np.uint16).reshape(3, 3, 1)
        headers = ututils.build_dummy_headers(volume.shape[2:], metadata)
        mv_no_headers = MedicalVolume(volume, self._AFFINE)
        mv_incorrect_headers = MedicalVolume(volume, self._AFFINE, headers={})
        mv = MedicalVolume(volume, self._AFFINE, headers=headers)

        mv_unchanged = mv_no_headers.apply_rescale()
        assert mv_unchanged.is_identical(mv_no_headers)

        mv_unchanged = mv_incorrect_headers.apply_rescale()
        assert mv_unchanged.is_identical(mv_incorrect_headers)

        assert np.allclose(mv.apply_rescale().volume, volume * 2.5 + 1.0)
        assert np.allclose(mv.apply_rescale(3, 1).volume, volume * 3.0 + 1.0)

        # test sync
        mv2 = mv.apply_rescale(dtype="single", sync=True)
        assert mv.get_metadata("RescaleSlope", dtype=float) == 2.5
        assert mv.get_metadata("RescaleIntercept", dtype=float) == 1.0

        assert mv2.get_metadata("RescaleSlope", default=None) is None
        assert mv2.get_metadata("RescaleIntercept", default=None) is None
        assert mv2.dtype == np.float32

    def test_apply_modality_lut(self):
        ds = pydicom.Dataset()
        lut = pydicom.Dataset()
        lut_data_little_endian = bytearray([10, 0, 11, 0, 12, 0, 13, 0])
        lut_data_big_endian = bytearray([0, 10, 0, 11, 0, 12, 0, 13])

        ds.PixelRepresentation = 1
        ds.ModalityLUTSequence = pydicom.Sequence()
        ds.ModalityLUTSequence.append(lut)

        lut.LUTDescriptor = [4, 2, 16]  # entries, first mapped, bits
        lut.LUTExplanation = "Test"
        ds.RescaleIntercept = "0"
        ds.RescaleSlope = "1"

        volume = np.arange(9, dtype=np.uint16).reshape(3, 3, 1)
        mv_no_headers = MedicalVolume(volume, self._AFFINE)
        mv_incorrect_headers = MedicalVolume(volume, self._AFFINE, headers=[pydicom.Dataset()])
        mv = MedicalVolume(volume, self._AFFINE, headers=[ds])

        mv_unchanged = mv_no_headers.apply_modality_lut()
        assert mv_unchanged.is_identical(mv_no_headers)

        mv_unchanged = mv_incorrect_headers.apply_modality_lut()
        assert mv_unchanged.is_identical(mv_incorrect_headers)

        # test modality lut
        correct = [10, 10, 10, 11, 12, 13, 13, 13, 13]
        correct_u8 = np.array(correct, dtype=np.uint8).reshape(3, 3, 1)

        # test big endian
        lut["LUTData"] = pydicom.DataElement("LUTData", "OW", lut_data_big_endian)
        mv_big_endian = mv.apply_modality_lut()
        assert np.allclose(mv_big_endian._volume, correct_u8)

        # test sync and inplace
        lut["LUTData"] = pydicom.DataElement("LUTData", "OW", lut_data_little_endian)
        mv_inplace = mv.apply_modality_lut(inplace=True, sync=True)
        assert mv_inplace is mv

    def test_apply_window(self):
        volume = np.arange(9, dtype=np.uint8)
        ds = pydicom.Dataset()
        ds.WindowCenter = "4.5"
        ds.WindowWidth = "4.0"
        ds.PhotometricInterpretation = "MONOCHROME2"
        ds.PixelRepresentation = 0
        ds.BitsAllocated = 8
        ds.BitsStored = 8
        ds.VOILUTFunction = "LINEAR"
        mv_no_headers = MedicalVolume(volume, self._AFFINE)
        mv_incorrect_headers = MedicalVolume(volume, self._AFFINE, headers={})
        mv = MedicalVolume(volume, self._AFFINE, headers=[ds])

        mv_unchanged = mv_no_headers.apply_window()
        assert mv_unchanged.is_identical(mv_no_headers)

        mv_unchanged = mv_incorrect_headers.apply_window()
        assert mv_unchanged.is_identical(mv_incorrect_headers)

        # test VOILUTFunction linear
        correct_linear = np.array([0, 0, 0, 42.5, 127.5, 212.5, 255, 255, 255])
        assert np.allclose(mv.apply_window(output_range=(0, 255)).volume, correct_linear)

        # test VOILUTFunction exact
        correct_exact = np.array([0.0, 0.0, 0.0, 42.5, 127.5, 212.5, 255.0, 255.0, 255.0])
        mv.set_metadata("VOILUTFunction", "LINEAR_EXACT", force=True)
        assert np.allclose(mv.apply_window(output_range=(0, 255)).volume, correct_exact)

    def test_apply_voi_lut(self):
        ds = pydicom.Dataset()
        lut1 = pydicom.Dataset()
        lut2 = pydicom.Dataset()
        lut_data_little_endian = bytearray([10, 11, 12, 13])
        lut_data_big_endian = bytearray([0, 10, 0, 11, 0, 12, 0, 13])

        ds.PixelRepresentation = 1
        ds.VOILUTSequence = pydicom.Sequence()
        ds.VOILUTSequence.extend([lut1, lut2])

        # VOILUTSequence can contain multiple LUTs
        lut1.LUTDescriptor = [4, 2, 16]  # entries, first mapped, bits
        lut1.LUTExplanation = "Test1"
        lut2.LUTDescriptor = [4, 2, 16]  # entries, first mapped, bits
        lut2.LUTExplanation = "Test2"

        ds.RescaleIntercept = "0"
        ds.RescaleSlope = "1"

        volume = np.arange(9, dtype=np.uint16).reshape(3, 3, 1)
        mv_no_headers = MedicalVolume(volume, self._AFFINE)
        mv_incorrect_headers = MedicalVolume(volume, self._AFFINE, headers=[pydicom.Dataset()])
        mv = MedicalVolume(volume, self._AFFINE, headers=[ds])

        mv_unchanged = mv_no_headers.apply_voi_lut()
        assert mv_unchanged.is_identical(mv_no_headers)

        mv_unchanged = mv_incorrect_headers.apply_voi_lut()
        assert mv_unchanged.is_identical(mv_incorrect_headers)

        # test modality lut1
        correct = [10, 10, 10, 11, 12, 13, 13, 13, 13]
        correct_u8 = np.array(correct, dtype=np.uint8).reshape(3, 3, 1)

        # test big endian
        lut1["LUTData"] = pydicom.DataElement("LUTData", "OW", lut_data_big_endian)
        mv_big_endian = mv.apply_voi_lut()
        assert np.allclose(mv_big_endian._volume, correct_u8)

        # test modality lut2
        lut2["LUTData"] = pydicom.DataElement("LUTData", "US", lut_data_little_endian)
        mv_lut2 = mv.apply_voi_lut(index=1)
        assert np.allclose(mv_lut2._volume, correct_u8)

    def test_clone(self):
        mv = MedicalVolume(np.random.rand(10, 20, 30), self._AFFINE)
        mv2 = mv.clone()
        assert mv.is_identical(mv2)  # expected identical volumes

        mv = MedicalVolume(
            np.random.rand(10, 20, 30),
            self._AFFINE,
            headers=ututils.build_dummy_headers((1, 1, 30)),
        )
        mv2 = mv.clone(headers=False)
        assert mv.is_identical(mv2)  # expected identical volumes
        assert id(mv.headers(flatten=True)[0]) == id(
            mv2.headers(flatten=True)[0]
        ), "headers not cloned, expected same memory address"

        mv3 = mv.clone(headers=True)
        assert mv.is_identical(mv3)  # expected identical volumes
        assert id(mv.headers(flatten=True)[0]) != id(
            mv3.headers(flatten=True)[0]
        ), "headers cloned, expected different memory address"

    def test_to_nib(self):
        arr = np.random.rand(10, 20, 30)
        mv = MedicalVolume(arr, self._AFFINE)
        nib_img = nib.Nifti1Image(arr, mv.affine)

        nib_from_mv = mv.to_nib()
        assert np.all(nib_from_mv.get_fdata() == nib_img.get_fdata())
        assert np.all(nib_from_mv.affine == nib_img.affine)

    def test_to_sitk(self):
        mv = MedicalVolume(np.random.rand(10, 20, 30), self._AFFINE)
        filepath = os.path.join(ututils.TEMP_PATH, "med_vol_to_sitk.nii.gz")
        NiftiWriter().save(mv, filepath)

        expected = sitk.ReadImage(filepath)

        nr = NiftiReader()
        mv = nr.load(filepath)
        img = mv.to_sitk()

        assert np.allclose(sitk.GetArrayViewFromImage(img), sitk.GetArrayViewFromImage(expected))
        assert img.GetSize() == mv.shape
        assert np.allclose(img.GetOrigin(), expected.GetOrigin())
        assert img.GetSpacing() == img.GetSpacing()
        assert img.GetDirection() == expected.GetDirection()

        mv = MedicalVolume(np.zeros((10, 20, 1, 3)), affine=self._AFFINE)
        img = mv.to_sitk(vdim=-1)
        assert np.all(sitk.GetArrayViewFromImage(img) == 0)
        assert img.GetSize() == (10, 20, 1)

        filepath = pydd.get_testdata_file("MR_small.dcm")
        dr = DicomReader(group_by=None)
        mv = dr.load(filepath)
        mv2 = MedicalVolume.from_sitk(
            mv.to_sitk(transpose_inplane=True), copy=True, transpose_inplane=True
        )
        assert mv2.is_identical(mv)

    @unittest.skipIf(not ututils.is_data_available(), "unittest data is not available")
    def test_to_from_sitk_dicom_convention(self):
        dp = ututils.get_scan_dirpath("qdess")
        dirpath = ututils.get_read_paths(dp, ImageDataFormat.dicom)[0]

        dr = DicomReader()
        dr = DicomReader(group_by=None, sort_by="InstanceNumber", ignore_ext=False)
        mv: MedicalVolume = dr.load(dirpath)[0]

        reader = sitk.ImageSeriesReader()
        dicom_names = reader.GetGDCMSeriesFileNames(dirpath)
        reader.SetFileNames(dicom_names)
        sitk_image = reader.Execute()

        sitk_from_mv = mv.to_sitk(transpose_inplane=True)
        img, expected = sitk_from_mv, sitk_image
        assert np.allclose(sitk.GetArrayViewFromImage(img), sitk.GetArrayViewFromImage(expected))
        assert img.GetSize() == mv.shape
        assert np.allclose(img.GetOrigin(), expected.GetOrigin())
        assert img.GetSpacing() == img.GetSpacing()
        assert img.GetDirection() == expected.GetDirection()

        mv_from_sitk = MedicalVolume.from_sitk(sitk_image, copy=True, transpose_inplane=True)
        assert mv_from_sitk.is_identical(mv)

        mv2 = MedicalVolume.from_sitk(sitk_from_mv, copy=True, transpose_inplane=True)
        assert mv2.is_identical(mv)

    def test_from_nib(self):
        filepath = os.path.join(nib_testing.data_path, "example4d.nii.gz")
        nib_img = nib.load(filepath)

        mv = MedicalVolume.from_nib(nib_img)
        assert np.all(mv.affine == nib_img.affine)
        assert np.all(mv.A == nib_img.get_fdata())

        precision = 4
        mv2 = MedicalVolume.from_nib(
            nib_img, affine_precision=precision, origin_precision=precision
        )
        assert np.allclose(mv2.affine, nib_img.affine, atol=10 ** (-precision))
        assert np.all(mv2.A == nib_img.get_fdata())

    def test_from_sitk(self):
        mv = MedicalVolume(np.random.rand(10, 20, 30), self._AFFINE)
        filepath = os.path.join(ututils.TEMP_PATH, "med_vol_from_sitk.nii.gz")
        NiftiWriter().save(mv, filepath)

        nr = NiftiReader()
        expected = nr.load(filepath)

        img = sitk.ReadImage(filepath)
        mv = MedicalVolume.from_sitk(img)

        assert np.allclose(mv.affine, expected.affine)
        assert mv.shape == expected.shape
        assert np.all(mv.volume == expected.volume)

        img = sitk.Image([10, 20, 1], sitk.sitkVectorFloat32, 3)
        mv = MedicalVolume.from_sitk(img)
        assert np.all(mv.volume == 0)
        assert mv.shape == (10, 20, 1, 3)

    def test_math(self):
        mv1 = MedicalVolume(np.ones((10, 20, 30)), self._AFFINE)
        mv2 = MedicalVolume(2 * np.ones((10, 20, 30)), self._AFFINE)

        out = mv1 + mv2
        assert np.all(mv1._volume == 1)
        assert np.all(mv2._volume == 2)
        assert np.all(out._volume == 3)

        out = mv1 - mv2
        assert np.all(mv1._volume == 1)
        assert np.all(mv2._volume == 2)
        assert np.all(out._volume == -1)

        out = mv1 * mv2
        assert np.all(mv1._volume == 1)
        assert np.all(mv2._volume == 2)
        assert np.all(out._volume == 2)

        out = mv1 / mv2
        assert np.all(mv1._volume == 1)
        assert np.all(mv2._volume == 2)
        assert np.all(out._volume == 0.5)

        out = mv1 // mv2
        assert np.all(mv1._volume == 1)
        assert np.all(mv2._volume == 2)
        assert np.all(out._volume == 0)

        out = mv1**mv2
        assert np.all(mv1._volume == 1)
        assert np.all(mv2._volume == 2)
        assert np.all(out._volume == 1)

        out = mv1.clone()
        out += mv2
        assert np.all(out._volume == 3)

        out = mv1.clone()
        out -= mv2
        assert np.all(out._volume == -1)

        out = mv1.clone()
        out *= mv2
        assert np.all(out._volume == 2)

        out = mv1.clone()
        out /= mv2
        assert np.all(out._volume == 0.5)

        out = mv1.clone()
        out //= mv2
        assert np.all(out._volume == 0)

        out = mv1.clone()
        out **= mv2
        assert np.all(out._volume == 1)

        mv3 = mv1.clone().reformat(mv1.orientation[::-1])
        with self.assertRaises(ValueError):
            mv3 + mv2

    def test_shape_attributes(self):
        vol = np.ones((10, 20, 30, 2))
        mv = MedicalVolume(vol, np.eye(4))
        assert mv.shape == vol.shape
        assert mv.ndim == vol.ndim

    def test_hdf5(self):
        shape = (10, 20, 30)
        volume = np.reshape(list(range(np.product(shape))), shape)
        hdf5_file = os.path.join(self._TEMP_PATH, "unittest.h5")

        with h5py.File(hdf5_file, "w") as f:
            f.create_dataset("volume", data=volume)
        f = h5py.File(hdf5_file, "r")

        mv = MedicalVolume(f["volume"], np.eye(4))
        assert mv.device == Device("cpu")
        assert mv.dtype == f["volume"].dtype

        mv2 = mv[:, :, :1]
        assert np.all(mv2.volume == volume[:, :, :1])
        assert mv2.device == Device("cpu")
        assert mv2.dtype == volume.dtype

    def test_comparison(self):
        mv1 = MedicalVolume(np.ones((10, 20, 30)), self._AFFINE)
        mv2 = MedicalVolume(2 * np.ones((10, 20, 30)), self._AFFINE)

        assert np.all((mv1 == mv1.clone()).volume)
        assert np.all((mv1 != mv2).volume)
        assert np.all((mv1 < mv2).volume)
        assert np.all((mv1 <= mv2).volume)
        assert np.all((mv2 > mv1).volume)
        assert np.all((mv2 >= mv1).volume)

    def test_slice(self):
        mv = MedicalVolume(np.ones((10, 20, 30)), self._AFFINE)
        with self.assertRaises(IndexError):
            mv[4]
        mv_slice = mv[4:5]
        assert mv_slice.shape == (1, 20, 30)

        mv = MedicalVolume(np.ones((10, 20, 30)), self._AFFINE)
        mv[:5, ...] = 2
        assert np.all(mv._volume[:5, ...] == 2) & np.all(mv._volume[5:, ...] == 1)
        assert np.all(mv[:5, ...].volume == 2)

        mv = MedicalVolume(np.ones((10, 20, 30)), self._AFFINE)
        mv2 = mv[:5, ...].clone()
        mv2 += 2
        mv[:5, ...] = mv2
        assert np.all(mv._volume[:5, ...] == 3) & np.all(mv._volume[5:, ...] == 1)
        assert np.all(mv[:5, ...].volume == 3)

    def test_slice_with_headers(self):
        vol = np.stack([np.ones((10, 20, 30)), 2 * np.ones((10, 20, 30))], axis=-1)
        headers = np.stack(
            [
                ututils.build_dummy_headers(vol.shape[2], {"EchoTime": 2}),
                ututils.build_dummy_headers(vol.shape[2], {"EchoTime": 10}),
            ],
            axis=-1,
        )
        mv = MedicalVolume(vol, self._AFFINE, headers=headers)

        mv2 = mv[..., 0]
        assert mv2._headers.shape == (1, 1, 30)
        for h in mv2.headers(flatten=True):
            assert h["EchoTime"].value == 2

        mv2 = mv[..., 1]
        assert mv2._headers.shape == (1, 1, 30)
        for h in mv2.headers(flatten=True):
            assert h["EchoTime"].value == 10

        mv2 = mv[:10, :5, 8:10, :1]
        assert mv2._headers.shape == (1, 1, 2, 1)

        mv2 = mv[:10]
        assert mv2._headers.shape == (1, 1, 30, 2)
        mv2 = mv[:, :10]
        assert mv2._headers.shape == (1, 1, 30, 2)

        mv2 = mv[..., 0:1]
        assert mv2._headers.shape == (1, 1, 30, 1)

        vol = np.stack([np.ones((10, 20, 30)), 2 * np.ones((10, 20, 30))], axis=-1)
        headers = ututils.build_dummy_headers(vol.shape[2], {"EchoTime": 2})[..., np.newaxis]
        mv = MedicalVolume(vol, self._AFFINE, headers=headers)
        mv1 = mv[..., 0]
        mv2 = mv[..., 1]
        assert mv1._headers.shape == (1, 1, 30)
        assert mv2._headers.shape == (1, 1, 30)
        for h1, h2 in zip(mv1.headers(flatten=True), mv2.headers(flatten=True)):
            assert id(h1) == id(h2)

    def test_4d(self):
        vol = np.stack([np.ones((10, 20, 30)), 2 * np.ones((10, 20, 30))], axis=-1)
        mv = MedicalVolume(vol, self._AFFINE)
        assert mv.orientation == ("SI", "AP", "LR")
        assert mv.shape == (10, 20, 30, 2)

        assert np.all(mv[..., 0].volume == 1)
        assert np.all(mv[..., 1].volume == 2)

        ornt = ("AP", "IS", "RL")
        mv2 = mv.reformat(ornt)
        assert mv2.orientation == ornt
        assert mv2.shape == (20, 10, 30, 2)

        mv2 = mv.reformat(ornt).reformat(mv.orientation)
        assert mv2.is_identical(mv)

        fp = os.path.join(self._TEMP_PATH, "test_4d.nii.gz")
        mv.save_volume(fp)
        mv2 = NiftiReader().load(fp)
        assert mv2.is_identical(mv)

    def test_device(self):
        mv = MedicalVolume(np.ones((10, 20, 30)), self._AFFINE)
        assert mv.to(Device(-1)) == mv
        assert mv.cpu() == mv

    @unittest.skipIf(not env.cupy_available(), "cupy not available")
    def test_device_gpu(self):
        import cupy as cp

        mv = MedicalVolume(np.ones((10, 20, 30)), self._AFFINE)
        mv_gpu = mv.to(Device(0))

        assert mv_gpu.device == Device(0)
        assert isinstance(mv_gpu.volume, cp.ndarray)
        assert isinstance(mv_gpu.affine, np.ndarray)

        assert mv_gpu.is_same_dimensions(mv)

        assert cp.all((mv_gpu + 1).volume == 2)
        assert cp.all((mv_gpu - 1).volume == 0)
        assert cp.all((mv_gpu * 2).volume == 2)
        assert cp.all((mv_gpu / 2).volume == 0.5)
        assert cp.all((mv_gpu > 0).volume)
        assert cp.all((mv_gpu >= 0).volume)
        assert cp.all((mv_gpu < 2).volume)
        assert cp.all((mv_gpu <= 2).volume)

        ornt = tuple(x[::-1] for x in mv_gpu.orientation[::-1])
        mv2 = mv_gpu.reformat(ornt)
        assert mv2.orientation == ornt

        mv_cpu = mv_gpu.cpu()
        assert mv_cpu.device == Device(-1)
        assert mv_cpu.is_identical(mv)

        with self.assertRaises(RuntimeError):
            mv_gpu.save_volume(os.path.join(self._TEMP_PATH, "test_device.nii.gz"))

    def test_array_cpu(self):
        mv = MedicalVolume(np.ones((10, 20, 30)), self._AFFINE)

        data = np.asarray(mv)
        assert np.shares_memory(data, mv.volume)

    @unittest.skipIf(not env.cupy_available(), "cupy not available")
    def test_array_gpu(self):
        import cupy as cp

        mv = MedicalVolume(np.ones((10, 20, 30)), self._AFFINE)
        mv_gpu = mv.to(Device(0))
        data = cp.asarray(mv_gpu)
        assert cp.shares_memory(data, mv_gpu.volume)

    def test_dtype(self):
        vol = np.ones((10, 20, 30))
        mv = MedicalVolume(vol, self._AFFINE)

        assert mv.volume.dtype == vol.dtype

        mv2 = mv.astype("int32")
        assert id(mv) == id(mv2)
        assert mv2.volume.dtype == np.int32

    def test_repr(self):
        vol = np.ones((10, 20, 30))
        mv = MedicalVolume(vol, self._AFFINE)

        assert mv.__repr__() is not None

    def test_set_volume(self):
        vol = np.ones((10, 20, 30))
        mv = MedicalVolume(vol, self._AFFINE)

        mv.volume += 2
        assert np.all(mv.volume == 3)

    @ututils.requires_packages("torch")
    def test_to_torch(self):
        import torch

        vol = np.ones((10, 20, 30))
        mv = MedicalVolume(vol, self._AFFINE)

        tensor = mv.to_torch()
        assert torch.all(tensor == torch.from_numpy(vol))
        assert tensor.shape == mv.shape

        tensor = mv.to_torch(requires_grad=True, contiguous=True)
        assert tensor.is_contiguous()
        assert tensor.requires_grad
        assert torch.all(tensor == torch.from_numpy(vol))
        assert tensor.shape == mv.shape

        vol = np.ones((10, 20, 30), np.complex)
        mv = MedicalVolume(vol, self._AFFINE)

        tensor = mv.to_torch()
        assert tensor.dtype == torch.complex128
        assert tensor.shape == mv.shape

        tensor = mv.to_torch(view_as_real=True)
        assert tensor.shape == mv.shape + (2,)

    @ututils.requires_packages("torch")
    def test_from_torch(self):
        import torch

        tensor = torch.ones(10, 20, 30)
        mv = MedicalVolume.from_torch(tensor, self._AFFINE)
        assert np.all(tensor.numpy() == mv.A)

        tensor = torch.ones(10, 20, 30)
        mv = MedicalVolume.from_torch(tensor, torch.from_numpy(self._AFFINE))
        assert np.all(tensor.numpy() == mv.A)
        assert isinstance(mv.affine, np.ndarray)

        tensor = torch.ones(10, 20, 30, dtype=torch.complex64)
        mv = MedicalVolume.from_torch(tensor, self._AFFINE)
        assert mv.dtype == np.complex64

        tensor = torch.ones(10, 20, 30, dtype=torch.complex128)
        mv = MedicalVolume.from_torch(tensor, self._AFFINE)
        assert mv.dtype == np.complex128

        tensor = torch.ones(10, 20, 30, 2, dtype=torch.float32)
        mv = MedicalVolume.from_torch(tensor, self._AFFINE, to_complex=True)
        assert mv.dtype == np.complex64
        assert mv.shape == tensor.shape[:3]

        tensor = torch.ones(10, 20, 30, 2, dtype=torch.float64)
        mv = MedicalVolume.from_torch(tensor, self._AFFINE, to_complex=True)
        assert mv.dtype == np.complex128
        assert mv.shape == tensor.shape[:3]

        tensor = torch.ones(10, 20, dtype=torch.float64)
        with self.assertRaises(ValueError):
            mv = MedicalVolume.from_torch(tensor, self._AFFINE)

        tensor = torch.ones(10, 20, 2, dtype=torch.float64)
        with self.assertRaises(ValueError):
            mv = MedicalVolume.from_torch(tensor, self._AFFINE, to_complex=True)

        tensor = torch.ones(10, 20, 30, 3, dtype=torch.float64)
        with self.assertRaises(ValueError):
            mv = MedicalVolume.from_torch(tensor, self._AFFINE, to_complex=True)

    @ututils.requires_packages("zarr")
    def test_to_zarr(self):
        import zarr

        vol = np.ones((10, 20, 30))
        mv = MedicalVolume(vol, self._AFFINE)
        arr = mv.to_zarr(read_only=True)

        assert np.array_equal(arr, vol)
        with self.assertRaises(zarr.errors.ReadOnlyError):
            arr[0, 0, 0] = 2

        aff = np.eye(4) * 2
        mv_aff = MedicalVolume(vol, aff)
        arr = mv_aff.to_zarr(affine_attr="affine")
        assert np.array_equal(np.array(arr.attrs["affine"]), aff)

        arr = mv_aff.to_zarr()
        with self.assertRaises(KeyError):
            arr.attrs["affine"]

        filepath = pydd.get_testdata_file("CT_small.dcm")
        dr = DicomReader()
        mv_hdr = dr.load(filepath)
        arr = mv_hdr.to_zarr(headers_attr="headers")
        assert "headers" in arr.attrs

    @ututils.requires_packages("zarr")
    def test_from_zarr(self):
        import zarr

        filepath = pydd.get_testdata_file("CT_small.dcm")
        dr = DicomReader()
        mv_hdr = dr.load(filepath)

        store = zarr.MemoryStore()
        arr = mv_hdr.to_zarr(store=store, affine_attr="affine", headers_attr="headers")
        assert arr.read_only

        # load without attributes
        mv = MedicalVolume.from_zarr(store=store)
        assert np.array_equal(mv.affine, np.eye(4))
        assert mv.headers() is None

        # load with attributes
        mv = MedicalVolume.from_zarr(store=store, affine_attr="affine", headers_attr="headers")
        assert np.array_equal(mv.affine, mv_hdr.affine)
        assert mv.headers() == mv_hdr.headers()

    def test_indexing(self):
        # Index medical volume with another medical volume.
        mv = MedicalVolume(np.ones((10, 20, 30)), np.eye(4))
        mv_index = MedicalVolume(np.random.rand(10, 20, 30) > 0.5, np.eye(4))
        mv[mv_index] = 0

        assert np.all(mv.A[mv_index.A] == 0)

    def test_mmap_numpy(self):
        shape = (5, 6, 7)

        # Memmap should directly write to the file in r+ mode.
        out_path = os.path.join(self._TEMP_PATH, "test.npy")
        np.save(out_path, np.ones(shape))

        arr = np.load(out_path, mmap_mode="r+")
        mv = MedicalVolume(arr, np.eye(4))
        assert mv.is_mmap

        mv += 1
        assert np.all(mv.A == 2)
        arr2 = np.load(out_path)
        assert np.all(arr2 == 2)

        # Memmap should be read-only.
        out_path = os.path.join(self._TEMP_PATH, "test2.npy")
        np.save(out_path, np.ones(shape))

        arr = np.load(out_path, mmap_mode="r")
        mv = MedicalVolume(arr, np.eye(4))
        assert mv.is_mmap

        orientation = mv.orientation
        mv = mv.reformat(orientation[::-1])
        assert mv.orientation == orientation[::-1]
        assert mv.shape == shape[::-1]

        mv2 = mv[1:3]
        assert np.all(mv2.A == mv.A[1:3])

        with self.assertRaises(ValueError):
            mv += 1

    def test_mmap_nifti(self):
        out_path = os.path.join(self._TEMP_PATH, "test.nii")
        mv = MedicalVolume(np.ones((10, 20, 30)), np.eye(4))
        mv.save_volume(out_path)

        vol = nib.load(out_path)
        assert isinstance(vol.dataobj.__array__(), np.memmap)
        mv2 = MedicalVolume.from_nib(vol, mmap=True)
        assert isinstance(mv2.A, np.memmap) and mv2.A.mode == "c"
        assert mv2.is_mmap
        assert mv2.is_identical(mv)

        mv2 += 1
        assert not mv2.is_mmap
        assert np.all(mv2.A == 2)

        mv2 = MedicalVolume.from_nib(vol, mmap=True)
        assert mv2.is_mmap
        mv2[1:4] = 0
        assert not mv2.is_mmap


if __name__ == "__main__":
    unittest.main()



"""
HTTP I/O.

This module contains HTTP input/output helpers.
"""

import re
import zipfile
from io import BytesIO
from typing import Collection, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

import numpy as np
import requests
from tqdm.auto import tqdm

from voxel.io.dicom import DicomReader
from voxel.io.format_io import DataReader, ImageDataFormat
from voxel.io.nifti import NiftiReader

__all__ = ["HttpReader"]


_MIME_TYPES_ZIP = [
    "application/zip",
    "application/x-zip-compressed",
    "multipart/x-zip",
    "application/dicom+zip",
]


class HttpReader(DataReader):
    """A class for reading DICOMs from HTTP requests with DICOMweb support.

    Attributes:
        verbose (bool, optional): If ``True``, show loading progress bar.
        block_size (int, optional): Block size for reading data.
        **kwargs: Keyword arguments for :class:`DicomReader`.

    Examples:
        >>> hr = HttpReader()
        >>> hr.read("https://server.com/dicom.zip")
        >>> hr.read("https://server.com/dicom.dcm")
        >>> hr.read("https://server.com/dicom-web/studies/x/series/y")
        >>> hr.close()

        >>> with HttpReader() as hr:
        >>>     hr.session.auth = ("username", "password")
        >>>     hr.read("https://server.com/dicom", params={"x": "y"})
    """

    def __init__(self, verbose: bool = False, block_size: int = 10**6, **kwargs):
        self.verbose = verbose
        self.block_size = block_size
        self.session = requests.Session()
        self.pbar = None
        self.kwargs = kwargs

    def _read_multipart_stream(
        self,
        res: requests.Response,
        content_info: str,
    ) -> List[bytes]:
        """Read multipart stream.

        Args:
            res (requests.Response): Response object.
            content_info (str): Content info.
            pbar (tqdm): Progress bar.
        """
        boundary = _extract_boundary(content_info)
        blob, parts = bytes(), []

        for block in res.iter_content(self.block_size):
            self.pbar.update(len(block))

            blob += block
            while boundary in blob:
                part, blob = blob.split(boundary, maxsplit=1)
                content = _extract_part(part)

                if content is not None:
                    parts.append(content)

        content = _extract_part(blob)
        if content is not None:
            parts.append(content)

        self.pbar.close()
        return parts

    def _read_stream(self, res: requests.Response) -> bytes:
        """Read stream.

        Args:
            res (requests.Response): Response object.
            pbar (tqdm): Progress bar.
        """

        blob = bytes()
        for block in res.iter_content(self.block_size):
            self.pbar.update(len(block))
            blob += block

        self.pbar.close()
        return blob

    def _read_dicom(self, buffers: List[bytes], **kwargs):
        """Read DICOMs from data.

        Args:
            buffers (List[bytes]): List of bytes objects.
            **kwargs: Keyword arguments for :class:`DicomReader`.
        """

        # do not pass verbose to reader, as files are already opened
        dr = DicomReader()
        return dr.read([BytesIO(buffer) for buffer in buffers], **kwargs)

    def _read_nifti(self, buffer: bytes, **kwargs):
        """Read NIfTI from data.

        Args:
            buffer (bytes): Bytes
            **kwargs: Keyword arguments for :class:`NiftiReader`.
        """
        nr = NiftiReader()
        return nr.read(BytesIO(buffer), **kwargs)

    def load(
        self,
        url: str,
        params: Union[Dict, List[Tuple], bytes] = np._NoValue,
        data_format: ImageDataFormat = None,
        verbose: bool = None,
        **kwargs,
    ):
        """Load data from HTTP request.

        Args:
            url (str): URL.
            params (Union[Dict, List[Tuple], bytes], optional): Parameters to send with the request.
            **kwargs: Keyword arguments for :class:`DicomReader`.
        """

        if not _is_valid_url(url):
            raise IOError(f"Invalid URL: {url}.")

        params = params if params != np._NoValue else self.session.params
        with self.session.get(url, params=params, stream=True) as res:
            content_length = res.headers.get("Content-Length", 0)
            content_type = res.headers.get("Content-Type", "application/octet-stream").lower()

            self.pbar = tqdm(
                total=content_length,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                disable=not verbose if verbose is not None else not self.verbose,
            )

            # Mime: multipart/related, expect DICOM
            if content_type.startswith("multipart/related;"):
                _, *content_info = [part.strip() for part in content_type.split(";")]
                parts = self._read_multipart_stream(res, content_info)
                return self._read_dicom(parts, **kwargs)

            # Mime: application/zip, expect DICOM
            if content_type in _MIME_TYPES_ZIP:
                blob = self._read_stream(res)
                z = zipfile.ZipFile(BytesIO(blob))
                parts = [z.read(zinfo) for zinfo in z.infolist() if zinfo.file_size > 0]
                return self._read_dicom(parts, **kwargs)

            # Fallback to single file, expect NiFTI or DICOM
            blob = self._read_stream(res)
            if data_format is None:
                basename = urlparse(url).path
                data_format = ImageDataFormat.get_image_data_format(basename)
            elif isinstance(data_format, str):
                data_format = ImageDataFormat[data_format]

            if data_format == ImageDataFormat.nifti:
                if urlparse(url).path.endswith(".gz"):
                    kwargs = {"compressed": True, **kwargs}
                return self._read_nifti(blob, **kwargs)
            elif data_format == ImageDataFormat.dicom:
                return self._read_dicom([blob], **kwargs)
            else:
                raise IOError(f"Unsupported data format: {data_format}.")

    def close(self):
        """Close the current HTTP session."""
        self.session.close()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def __serializable_variables__(self) -> Collection[str]:
        return self.__dict__.keys()

    read = load  # pragma: no cover


def _is_valid_url(url: str) -> bool:
    """Check if a string represents a valid URL.

    Args:
        url (str): URL.

    Returns:
        bool: Result of the URL validation.
    """
    regex = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
        r"localhost|"  # localhost...
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )

    return re.match(regex, url) is not None


def _extract_part(part: bytes) -> Union[bytes, None]:
    """Extract part from multipart stream."""
    if part in [b"", b"--", b"\r\n"] or part.startswith(b"--\r\n"):
        return None

    idx = part.index(b"\r\n\r\n")
    if idx > -1:
        return part[idx + 4 :]

    raise ValueError("Part is not CRLF CRLF terminated.")


def _extract_boundary(content_info: List[str]) -> Optional[bytes]:
    """Extract boundary from content info."""
    for item in content_info:
        if "=" not in item:
            continue

        key, value = item.split("=", maxsplit=1)
        if key.lower() == "boundary":
            return b"--" + value.strip('"').encode("utf-8")

    return None



"""NIfTI I/O.

This module contains NIfTI input/output helpers.
"""

import gzip
import os
from io import BytesIO
from typing import Collection, Union

import nibabel as nib

import voxel as vx
from voxel.io.format_io import DataReader, DataWriter, ImageDataFormat
from voxel.med_volume import MedicalVolume

__all__ = ["NiftiReader", "NiftiWriter"]


class NiftiReader(DataReader):
    """A class for reading NIfTI files.

    Attributes:
        data_format_code (ImageDataFormat): The supported image data format.
    """

    data_format_code = ImageDataFormat.nifti

    def load(
        self,
        path_or_bytes: Union[str, bytes, os.PathLike, BytesIO],
        mmap: bool = False,
        compressed: bool = False,
    ) -> MedicalVolume:
        """Load volume from NIfTI file path.

        A NIfTI file should only correspond to one volume.

        Args:
            path_or_bytes (Union[str, bytes, os.PathLike, BytesIO]): Path to NIfTI file or
                bytes of NIfTI file.
            mmap (bool): Whether to use memory mapping.
            compressed (bool): Whether to apply gzip decompression. This is only used if
                `path_or_bytes` is a bytes or BytesIO object.

        Returns:
            MedicalVolume: Loaded volume.

        Raises:
            FileNotFoundError: If `file_path` not found.
            ValueError: If `file_path` does not end in a supported NIfTI extension.
        """
        if isinstance(path_or_bytes, bytes):
            path_or_bytes = BytesIO(path_or_bytes)

        if isinstance(path_or_bytes, BytesIO):
            if compressed:
                path_or_bytes = BytesIO(gzip.decompress(path_or_bytes.getvalue()))

            nifti_version = _nifti_version(path_or_bytes)
            fh = nib.FileHolder(fileobj=path_or_bytes)
            if nifti_version == 1:
                nib_img = nib.Nifti1Image.from_file_map({"header": fh, "image": fh})
            else:
                nib_img = nib.Nifti2Image.from_file_map({"header": fh, "image": fh})

        else:
            if not os.path.isfile(path_or_bytes):
                raise FileNotFoundError("{} not found".format(path_or_bytes))

            if not self.data_format_code.is_filetype(path_or_bytes):
                raise ValueError(
                    "{} must be a file with extension '.nii' or '.nii.gz'".format(path_or_bytes)
                )

            nib_img = nib.load(path_or_bytes)

        return MedicalVolume.from_nib(
            nib_img,
            affine_precision=vx.config.affine_precision,
            origin_precision=vx.config.affine_precision,
            mmap=mmap,
        )

    def __serializable_variables__(self) -> Collection[str]:
        return self.__dict__.keys()

    read = load  # pragma: no cover


class NiftiWriter(DataWriter):
    """A class for writing volumes in NIfTI format.

    Attributes:
        data_format_code (ImageDataFormat): The supported image data format.
    """

    data_format_code = ImageDataFormat.nifti

    def save(self, volume: MedicalVolume, file_path: str):
        """Save volume in NIfTI format,

        Args:
            volume (MedicalVolume): Volume to save.
            file_path (str): File path to NIfTI file.

        Raises:
            ValueError: If `file_path` does not end in a supported NIfTI extension.
        """
        if not self.data_format_code.is_filetype(file_path):
            raise ValueError(
                "{} must be a file with extension '.nii' or '.nii.gz'".format(file_path)
            )

        # Create dir if does not exist
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        nib_img = volume.to_nib()
        nib.save(nib_img, file_path)

    def __serializable_variables__(self) -> Collection[str]:
        return self.__dict__.keys()

    write = save  # pragma: no cover


def _nifti_version(buffer: BytesIO) -> int:
    """Get NIfTI version from buffer."""
    nii1_sizeof_hdr = 348
    nii2_sizeof_hdr = 540

    byte_data = buffer.read(4)
    sizeof_hdr = int.from_bytes(byte_data, byteorder="little")
    if sizeof_hdr == nii1_sizeof_hdr:
        return 1
    elif sizeof_hdr == nii2_sizeof_hdr:
        return 2
    else:
        sizeof_hdr = int.from_bytes(byte_data, byteorder="big")
        if sizeof_hdr == nii1_sizeof_hdr:
            return 1
        elif sizeof_hdr == nii2_sizeof_hdr:
            return 2

    raise ValueError(
        "This buffer is not a valid NIfTI file. Pass `compressed=True` if the buffer is gzipped."
    )



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path

import yaml

CONFIG_ENV_VARIABLE = "VOXEL_CONFIG"


@dataclass
class VoxelConfig:

    # The precision to use for the affine matrix.
    affine_precision: int = 4
    # The default format to save MedicalVolume objects.
    save_format: str = "nifti"

    @classmethod
    def from_yaml(cls, path: str = None):
        if path is None:
            path = os.environ.get(
                CONFIG_ENV_VARIABLE,
                os.path.abspath(
                    os.path.join(os.path.join(Path.home(), ".cache", "voxel"), "config.yaml")
                ),
            )
            if not os.path.exists(path):
                # create empty config
                os.makedirs(os.path.dirname(path), exist_ok=True)
                yaml.dump({}, open(path, "w"))
        config = yaml.load(open(path, "r"), Loader=yaml.FullLoader)

        return cls(**config)


config = VoxelConfig.from_yaml()



import os
import sys

import nibabel
import numpy as np
import pydicom
from tabulate import tabulate

__all__ = ["collect_env_info"]


def collect_env_info():
    """Collect environment information for reporting issues.

    Run this function when reporting issues on Github.
    """
    data = []
    data.append(("sys.platform", sys.platform))
    data.append(("Python", sys.version.replace("\n", "")))
    data.append(("numpy", np.__version__))

    try:
        import voxel  # noqa

        data.append(("voxel", voxel.__version__ + " @" + os.path.dirname(voxel.__file__)))
    except ImportError:
        data.append(("voxel", "failed to import"))

    # Required packages
    data.append(("nibabel", nibabel.__version__))
    data.append(("pydicom", pydicom.__version__))

    # Optional packages
    try:
        import cupy

        data.append(("cupy", cupy.__version__))
    except ImportError:
        pass

    try:
        import sigpy

        data.append(("sigpy", sigpy.__version__))
    except ImportError:
        pass

    env_str = tabulate(data)
    return env_str



"""Functions and classes for getting and setting computing devices."""
import numpy as np

from voxel.utils import env

if env.cupy_available():
    import cupy as cp
if env.sigpy_available():
    import sigpy as sp
if env.torch_available():
    import torch

__all__ = ["Device", "get_device", "to_device"]


class Device(object):
    """Device class.

    This class extends ``cupy.Device`` and can also be used to interface with
    ``torch.Device`` and ``sigpy.Device``. This class contains a device type
    ('cpu' or 'cuda') and optional device ordinal (i.e. the id) for the device type.
    This class can also be constructed using only the ordinal id, where id >= 0
    representing the id-th GPU, and id = -1 representing CPU. cupy must be installed to use GPUs.

    The array module for the corresponding device can be obtained via ``device.xp``.
    Similar to cupy.Device, the Device object can be used as a context:

        >>> device = Device(1)  # gpu 1
        >>> xp = device.xp  # xp is cupy.
        >>> with device:
        >>>     x = xp.array([1, 2, 3])
        >>>     x += 1


    Args:
        id_or_device (int or Device or cupy.cuda.Device): id > 0 represents
            the corresponding GPUs, and id = -1 represents CPU.

    Attributes:
        type (str): Device type. Either ``'cpu'`` or ``'cuda'``.
        index (int): index = -1 represents CPU, and others represents the id_th ordinances.

    Note:
        This class is heavily based on
        `sigpy.Device <https://sigpy.readthedocs.io/en/latest/generated/sigpy.Device.html>`_.
    """

    def __init__(self, id_or_device):
        _type, id = None, None
        if isinstance(id_or_device, int):
            id = id_or_device
        elif id_or_device == "cpu":
            _type, id = id_or_device, -1
        elif id_or_device == "cuda":
            _type, id = id_or_device, 0
        elif isinstance(id_or_device, Device):
            _type, id = id_or_device.type, id_or_device.id
        elif env.cupy_available() and isinstance(id_or_device, cp.cuda.Device):
            _type, id = "cuda", id_or_device.id
        elif env.sigpy_available() and isinstance(id_or_device, sp.Device):
            id = id_or_device.id
        elif env.torch_available() and isinstance(id_or_device, torch.device):
            _type, id = id_or_device.type, id_or_device.index
            if id is None:
                if _type == "cuda":
                    id = torch.cuda.current_device()
                elif _type == "cpu":
                    id = -1
                else:
                    raise ValueError(f"Unsupported device type: {_type}")
        else:
            raise ValueError(
                f"Accepts int, Device, cupy.cuda.Device, or torch.device" f"got {id_or_device}"
            )

        assert id >= -1
        if _type is None:
            _type = "cpu" if id == -1 else "cuda"

        cpdevice = None
        if id != -1:
            if env.cupy_available():
                cpdevice = cp.cuda.Device(id)
            else:
                raise ValueError("cupy not installed, but set device {}.".format(id))

        self._type = _type
        self._id = id
        self._cpdevice = cpdevice

    @property
    def id(self):
        """int: The device ordinal."""
        return self._id

    @property
    def type(self):
        """str: Type of device. Either ``"cpu"`` or ``"cuda"``."""
        return self._type

    @property
    def index(self):
        """int: Alias for ``self.id``."""
        return self.id

    @property
    def cpdevice(self):
        """cupy.Device: The equivalent ```cupy.Device```."""
        return self._cpdevice

    @property
    def ptdevice(self):
        """torch.device: The equivalent ```torch.device```."""
        if not env.torch_available():
            raise RuntimeError("`torch` not installed.")

        if self.id == -1:
            return torch.device("cpu")

        return torch.device(f"{self.type}:{self.id}")

    @property
    def spdevice(self):
        """sigpy.Device: The equivalent ```sigpy.Device```."""
        if not env.sigpy_available():
            raise RuntimeError("`sigpy` not installed.")
        if self.id >= 0 and self.type != "cuda":
            raise RuntimeError(f"sigpy.Device does not support type {self.type}")
        return sp.Device(self.id)

    @property
    def xp(self):
        """module: numpy or cupy module for the device."""
        if self.id == -1:
            return np
        return cp

    def use(self):
        """Use computing device.

        All operations after use() will use the device when using ``cupy``.
        """
        if self.id > 0:
            self.cpdevice.use()

    def __int__(self):
        return self.id

    def __eq__(self, other):
        if other == -1:
            # other integers are not compared as self.type may be subject to change
            return self.id == other
        elif isinstance(other, Device):
            return self.type == other.type and self.id == other.id
        elif env.cupy_available() and isinstance(other, cp.cuda.Device):
            return self.type == "cuda" and self.id == other.id
        elif env.sigpy_available() and isinstance(other, sp.Device):
            try:
                return self.spdevice == other
            except RuntimeError:
                return False
        elif env.torch_available() and isinstance(other, torch.device):
            return self.ptdevice == other
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def __enter__(self):
        if self.id == -1:
            return None

        return self.cpdevice.__enter__()

    def __exit__(self, *args):
        if self.id == -1:
            pass
        else:
            self.cpdevice.__exit__()

    def __repr__(self):
        if self.id == -1:
            return "Device(type='cpu')"

        return f"Device(type='{self.type}', index={self.id})"


cpu_device = Device(-1)


def get_array_module(array):
    """Gets an appropriate module from :mod:`numpy` or :mod:`cupy`.

    This is almost equivalent to :func:`cupy.get_array_module`. The differences
    are that this function can be used even if cupy is not available.

    Adapted from :mod:`sigpy`.

    Args:
        array: Input array.

    Returns:
        module: :mod:`cupy` or :mod:`numpy` is returned based on input.
    """
    if env.cupy_available():
        return cp.get_array_module(array)
    else:
        return np


def get_device(array):
    """Get Device from input array.

    Adapted from :mod:`sigpy`.

    Args:
        array (array): Array.

    Returns:
        Device.
    """
    if hasattr(array, "device"):
        return Device(array.device)
    else:
        return cpu_device


def to_device(input, device=cpu_device):
    """Move input to device. Does not copy if same device.

    Adapted from :mod:`sigpy`.

    Args:
        input (array): Input.
        device (int or Device or cupy.Device): Output device.

    Returns:
        array: Output array placed in device.
    """
    idevice = get_device(input)
    odevice = Device(device)

    if idevice == odevice:
        return input

    if odevice == cpu_device:
        with idevice:
            return input.get()
    else:
        with odevice:
            return cp.asarray(input)



import unittest

import numpy as np

from voxel.io.format_io import ImageDataFormat
from voxel.io.nifti import NiftiReader, NiftiWriter
from voxel.med_volume import MedicalVolume
from voxel.orientation import to_affine

from . import util as ututils


class TestOrientation(unittest.TestCase):
    nr = NiftiReader()
    nw = NiftiWriter()

    def setUp(self):
        # filepaths to all separated echos
        # (done in image-viewer such as ITK-Snap, Horos, etc)
        filepaths = []
        for fp in ututils.SCAN_DIRPATHS:
            filepaths.extend(ututils.get_read_paths(fp, ImageDataFormat.nifti))
        self.filepaths = filepaths

    def check_orientations(self, mv: MedicalVolume, orientations):
        """Check the orientation of the medical volume is consistent.

        Apply each orientation specified in orientations to the Medical Volume mv Assert if mv
        --> apply orientation --> apply original orientation != mv original position coordinates.

        Args:
            mv: a Medical Volume
            orientations: a list or tuple of orientation tuples
        """
        o_base, so_base, ps_base = mv.orientation, mv.scanner_origin, mv.pixel_spacing
        ps_affine = np.array(mv.affine)

        for o in orientations:
            # Reorient to some orientation
            mv.reformat(o, inplace=True)

            # Reorient to original orientation
            mv.reformat(o_base, inplace=True)

            assert mv.orientation == o_base, "Orientation mismatch: Expected %s, got %s" % (
                str(o_base),
                str(mv.orientation),
            )
            assert mv.scanner_origin == so_base, "Scanner Origin mismatch: Expected %s, got %s" % (
                str(so_base),
                str(mv.scanner_origin),
            )
            assert mv.pixel_spacing == ps_base, "Pixel Spacing mismatch: Expected %s, got %s" % (
                str(ps_base),
                str(mv.pixel_spacing),
            )

            assert (
                mv.affine == ps_affine
            ).all(), "Affine matrix mismatch: Expected\n%s\ngot\n%s" % (
                str(ps_affine),
                str(mv.affine),
            )

    @unittest.skipIf(not ututils.is_data_available(), "unittest data is not available")
    def test_flip(self):
        # tests flipping orientation across volume axis
        for fp in self.filepaths:
            e1 = self.nr.load(fp)
            o = e1.orientation
            orientations = [
                (o[0][::-1], o[1], o[2]),
                (o[0], o[1][::-1], o[2]),
                (o[0], o[1][::-1], o[2]),
                (o[0][::-1], o[1][::-1], o[2]),
                (o[0][::-1], o[1], o[2][::-1]),
                (o[0], o[1][::-1], o[2][::-1]),
                (o[0][::-1], o[1][::-1], o[2][::-1]),
            ]

            self.check_orientations(e1, orientations)

    @unittest.skipIf(not ututils.is_data_available(), "unittest data is not available")
    def test_transpose(self):
        # tests transposing axes - i.e. changing vantage point on current volume
        for fp in self.filepaths:
            e1 = self.nr.load(fp)
            o = e1.orientation

            orientations = [
                (o[1], o[2], o[0]),
                (o[2], o[0], o[1]),
                (o[1], o[0], o[2]),
                (o[2], o[1], o[0]),
                (o[0], o[2], o[1]),
            ]

            self.check_orientations(e1, orientations)

    @unittest.skipIf(not ututils.is_data_available(), "unittest data is not available")
    def test_transpose_and_flip(self):
        from itertools import permutations

        for fp in self.filepaths:
            e1 = self.nr.load(fp)
            o = e1.orientation

            # generate all possible transposed versions of the existing orientation
            transpose_orientations = list(permutations(o))
            flip_orientations_indices = [[0], [1], [2], [0, 1], [0, 2], [1, 2], [0, 1, 2]]
            orientations = []
            for to in transpose_orientations:
                for fo_inds in flip_orientations_indices:
                    o_test = list(to)
                    for i in fo_inds:
                        o_test[i] = o_test[i][::-1]
                    orientations.append(tuple(o_test))

            self.check_orientations(e1, orientations)


class TestToAffine(unittest.TestCase):
    """Test cases for voxel.orientation.to_affine."""

    def test_basic(self):
        """Basic transposes and flips in RAS+ coordinate system."""
        orientations = [
            ("LR", "PA", "IS"),  # standard RAS+
            ("RL", "AP", "SI"),  # flipped
            ("IS", "LR", "PA"),  # transposed
            ("AP", "SI", "RL"),  # transposed + flipped
        ]

        for ornt in orientations:
            affine = to_affine(ornt)
            mv = MedicalVolume(np.ones((10, 20, 30)), affine)
            assert mv.orientation == ornt
            assert np.all(np.asarray(mv.pixel_spacing) == 1)
            assert np.all(np.asarray(mv.scanner_origin) == 0)

    def test_spacing(self):
        """Test affine matrix with pixel spacing."""
        ornt = ("AP", "SI", "RL")

        spacing = np.random.rand(3) + 0.1  # avoid pixel spacing of 0
        affine = to_affine(ornt, spacing)
        mv = MedicalVolume(np.ones((10, 20, 30)), affine)
        assert mv.orientation == ornt
        assert np.all(np.asarray(mv.pixel_spacing) == spacing)
        assert np.all(np.asarray(mv.scanner_origin) == 0)

        spacing = np.random.rand(1) + 0.1  # avoid pixel spacing of 0
        expected_spacing = np.asarray(list(spacing) + [1.0, 1.0])
        affine = to_affine(ornt, spacing)
        mv = MedicalVolume(np.ones((10, 20, 30)), affine)
        assert mv.orientation == ornt
        assert np.all(np.asarray(mv.pixel_spacing) == expected_spacing)
        assert np.all(np.asarray(mv.scanner_origin) == 0)

    def test_origin(self):
        """Test affine matrix with scanner origin."""
        ornt = ("AP", "SI", "RL")

        origin = np.random.rand(3)
        affine = to_affine(ornt, spacing=None, origin=origin)
        mv = MedicalVolume(np.ones((10, 20, 30)), affine)
        assert mv.orientation == ornt
        assert np.all(np.asarray(mv.pixel_spacing) == 1)
        assert np.all(np.asarray(mv.scanner_origin) == origin)

        origin = np.random.rand(1)
        expected_origin = np.asarray(list(origin) + [0.0, 0.0])
        affine = to_affine(ornt, spacing=None, origin=origin)
        mv = MedicalVolume(np.ones((10, 20, 30)), affine)
        assert mv.orientation == ornt
        assert np.all(np.asarray(mv.pixel_spacing) == 1)
        assert np.all(np.asarray(mv.scanner_origin) == expected_origin)

    def test_complex(self):
        ornt = ("AP", "SI", "RL")
        spacing = (0.5, 0.7)
        origin = (100, -54)

        expected_spacing = np.asarray(list(spacing) + [1.0])
        expected_origin = np.asarray(list(origin) + [0.0])

        affine = to_affine(ornt, spacing=spacing, origin=origin)
        mv = MedicalVolume(np.ones((10, 20, 30)), affine)
        assert mv.orientation == ornt
        assert np.all(np.asarray(mv.pixel_spacing) == expected_spacing)
        assert np.all(np.asarray(mv.scanner_origin) == expected_origin)


if __name__ == "__main__":
    unittest.main()



"""List of numpy functions supported for MedicalVolumes."""
import warnings
from typing import Sequence, Union

import numpy as np

from voxel.device import cpu_device
from voxel.med_volume import MedicalVolume

__all__ = [
    "amin",
    "amax",
    "argmin",
    "argmax",
    "sum_np",
    "mean_np",
    "std",
    "nanmin",
    "nanmax",
    "nanargmin",
    "nanargmax",
    "nansum",
    "nanmean",
    "nanstd",
    "nan_to_num",
    "around",
    "clip",
    "stack",
    "concatenate",
    "expand_dims",
    "squeeze",
    "pad",
    "where",
    "all_np",
    "any_np",
    "zeros_like",
    "ones_like",
    "shares_memory",
    "may_share_memory",
]


_HANDLED_NUMPY_FUNCTIONS = {}


def implements(*np_functions):
    "Register an __array_function__ implementation for DiagonalArray objects."

    def decorator(func):
        for np_func in np_functions:
            _HANDLED_NUMPY_FUNCTIONS[np_func] = func
        return func

    return decorator


def reduce_array_op(func, x, axis=None, **kwargs):
    kwargs = {k: v for k, v in kwargs.items() if v != np._NoValue}
    input = x._extract_input_array_ufunc(x)
    if input is NotImplemented:
        return NotImplemented
    return x._reduce_array(func, input, axis=axis, **kwargs)


@implements(np.amin)
def amin(x, axis=None, keepdims=False, initial=np._NoValue, where=np._NoValue):
    """See :func:`numpy.amin`."""
    return reduce_array_op(np.amin, x, axis=axis, keepdims=keepdims, initial=initial, where=where)


@implements(np.amax)
def amax(x, axis=None, keepdims=False, initial=np._NoValue, where=np._NoValue):
    """See :func:`numpy.amax`."""
    return reduce_array_op(np.amax, x, axis=axis, keepdims=keepdims, initial=initial, where=where)


@implements(np.argmin)
def argmin(x, axis=None):
    """See :func:`numpy.argmin`."""
    return reduce_array_op(np.argmin, x, axis=axis)


@implements(np.argmax)
def argmax(x, axis=None):
    """See :func:`numpy.argmax`."""
    return reduce_array_op(np.argmax, x, axis=axis)


@implements(np.sum)
def sum_np(x, axis=None, dtype=None, keepdims=False, initial=np._NoValue, where=np._NoValue):
    """See :func:`numpy.sum`."""
    return reduce_array_op(
        np.sum, x, axis=axis, dtype=dtype, keepdims=keepdims, initial=initial, where=where
    )


@implements(np.mean)
def mean_np(x, axis=None, dtype=None, keepdims=False, where=np._NoValue):
    """See :func:`numpy.mean`."""
    return reduce_array_op(np.mean, x, axis=axis, dtype=dtype, keepdims=keepdims, where=where)


@implements(np.std)
def std(x, axis=None, dtype=None, ddof=0, keepdims=False, where=np._NoValue):
    """See :func:`numpy.std`."""
    return reduce_array_op(
        np.std, x, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims, where=where
    )


@implements(np.nanmin)
def nanmin(x, axis=None, keepdims=False):
    """See :func:`numpy.nanmin`."""
    return reduce_array_op(np.nanmin, x, axis=axis, keepdims=keepdims)


@implements(np.nanmax)
def nanmax(x, axis=None, keepdims=False):
    """See :func:`numpy.nanmax`."""
    return reduce_array_op(np.nanmax, x, axis=axis, keepdims=keepdims)


@implements(np.nanargmin)
def nanargmin(x, axis=None):
    """See :func:`numpy.nanargmin`."""
    return reduce_array_op(np.nanargmin, x, axis=axis)


@implements(np.nanargmax)
def nanargmax(x, axis=None):
    """See :func:`numpy.nanargmax`."""
    return reduce_array_op(np.nanargmax, x, axis=axis)


@implements(np.nansum)
def nansum(x, axis=None, dtype=None, keepdims=False):
    """See :func:`numpy.nansum`."""
    return reduce_array_op(np.nansum, x, axis=axis, dtype=dtype, keepdims=keepdims)


@implements(np.nanmean)
def nanmean(x, axis=None, dtype=None, keepdims=False):
    """See :func:`numpy.nanmean`."""
    return reduce_array_op(np.nanmean, x, axis=axis, dtype=dtype, keepdims=keepdims)


@implements(np.nanstd)
def nanstd(x, axis=None, dtype=None, ddof=0, keepdims=False):
    """See :func:`numpy.nanstd`."""
    return reduce_array_op(np.nanstd, x, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims)


@implements(np.nan_to_num)
def nan_to_num(x, copy=True, nan=0.0, posinf=None, neginf=None):
    """See :func:`numpy.nan_to_num`."""
    vol = np.nan_to_num(x.volume, copy=copy, nan=nan, posinf=posinf, neginf=neginf)
    if not copy:
        x._volume = vol
        return x
    else:
        return x._partial_clone(volume=vol)


@implements(np.around, np.round, np.round_)
def around(x, decimals=0, affine=False):
    """Round medical image pixel data (and optionally affine) to the given number of decimals.

    Args:
        x (MedicalVolume): A medical image.
        decimals (int, optional): Number of decimal places to round to.
            If decimals is negative, it specifies the number of positions to the left
            of the decimal point.
        affine (bool, optional): If ``True``, rounds affine matrix.

    Returns:
        MedicalVolume: The rounded medical image.

    Examples:
        >>> mv = MedicalVolume(10*np.random.rand(3,4,5), affine=np.eye(4))
        >>> mv_rounded = np.round(mv, decimals=3)
    """
    affine = np.around(x.affine, decimals=decimals) if affine else x.affine
    return x._partial_clone(volume=np.around(x.volume, decimals=decimals), affine=affine)


@implements(np.clip)
def clip(x, x_min, x_max, **kwargs):
    """Clip the values in the array.

    Same as applying :func:`numpy.clip` on ``x.volume``.
    Only one of ``x_min`` or ``x_max`` can be ``None``.

    Args:
        x (MedicalVolume): Medical image to clip.
        x_min (array-like or ``MedicalVolume``): Minimum value.
            If ``None``, clipping is not performed on this edge.
        x_max (array-like or ``MedicalVolume``): Maximum value.
            If ``None``, clipping is not performed on this edge.
        kwargs: Optional keyword arguments, see :func:`numpy.clip`.

    Returns:
        MedicalVolume: The clipped medical image.

    Note:
        The ``out`` positional argument is not currently supported.

    Examples:
        >>> mv = MedicalVolume([[[0,1,2,3,4,5,6,7,8,9]]], affine=np.eye(4))
        >>> np.clip(mv, 1, 5)  # Clip values between [1, 5]
        MedicalVolume(volume=[[[1 1 2 3 4 5 5 5 5 5]]])
        >>> np.clip(mv, x_max=5)  # Clip values between (-inf, 5]
        MedicalVolume(volume=[[[0 1 2 3 4 5 5 5 5 5]]])
    """
    if isinstance(x_min, MedicalVolume):
        x_min = x_min.reformat_as(x).A
    if isinstance(x_max, MedicalVolume):
        x_max = x_max.reformat_as(x).A

    arr = np.clip(x.A, x_min, x_max, **kwargs)
    return x._partial_clone(volume=arr)


@implements(np.stack)
def stack(xs, axis: int = -1):
    """Stack medical images across non-spatial dimensions.

    Images will be auto-oriented to the orientation of the first medical volume.

    Args:
        xs (array-like[MedicalVolume]): 1D array-like of aligned medical images to stack.
        axis (int, optional): Axis to stack along.

    Returns:
        MedicalVolume: The stacked medical image.

    Note:
        Unlike NumPy, the default stacking axis is ``-1``.

    Note:
        Headers are not set unless all inputs have headers of the same
        shape. This functionality may change in the future.

    Examples:
        >>> mv = vx.MedicalVolume([[[0,1,2,3]]], affine=np.eye(4))
        >>> np.stack([mv, mv], axis=-1)  # Stacks along last axis.
        MedicalVolume(volume=[[[[0 0], [1 1], [2 2], [3 3]]]])
    """
    if not isinstance(axis, int):
        raise TypeError(f"'{type(axis)}' cannot be interpreted as int")

    xs = [x.reformat(xs[0].orientation) for x in xs]
    affine = xs[0].affine
    for x in xs[1:]:
        assert x.is_same_dimensions(xs[0], err=True)
    try:
        axis = _to_positive_axis(axis, len(xs[0].shape), grow=True, invalid_axis="spatial")
    except ValueError:
        raise ValueError(f"Cannot stack across spatial dimension (axis={axis})")
    assert axis >= 0

    vol = np.stack([x.volume for x in xs], axis=axis)
    headers = [x.headers() for x in xs]
    if any(x is None for x in headers):
        headers = None
    else:
        headers = np.stack(headers, axis=axis)

    return MedicalVolume(vol, affine, headers=headers)


@implements(np.concatenate)
def concatenate(xs, axis: int = -1):
    """Concatenate medical images.

    Image concatenation is slightly different if the axis is a spatial axis
    (one of the first 3 dimensions) or a non-spatial dimension.

    If concatenating along a non-spatial dimension, the image dimensions for all
    other axes and affine matrix of each ``x`` must be the same, which is standard
    for concatenation.

    If concatenating along a spatial dimension, all images must have the same direction
    and pixel spacing. Additionally, the scanner origin for all spatial axes not being
    concatenated should be the same. The origin for other scans should be consecutive.
    For example, if images are concatenated on ``axis=i``, a spatial axis, then
    ``xs[0].scanner_origin + xs[0].``.

    Images will be auto-oriented to the orientation of the first medical volume.

    Args:
        xs (Sequence[MedicalVolume]): The medical images to concatenate.
        axis (int, optional): The axis to concatenate on.

    Returns:
        MedicalVolume: The concatenated medical image.

    Note:
        Headers are not set unless all inputs have headers of the same
        shape. This functionality may change in the future.

    Examples:
        >>> mv = vx.MedicalVolume([[[[0],[1],[2],[3]]]], affine=np.eye(4))
        >>> np.concatenate([mv, mv], axis=-1)  # Concatenate along non-spatial dimension
        MedicalVolume(volume=[[[[0 0], [1 1], [2 2], [3 3]]]])
        >>> mv2 = vx.MedicalVolume(
            [[[[4],[5],[6],[7]]]],
            affine=[[1,0,0,0], [0,1,0,0],[0,0,1,4],[0,0,0,1]]
        )
        >>> np.concatenate([mv, mv2], axis=2)  # Concatenate along spatial dimension
        MedicalVolume(volume=[[[[0]
            [1]
            [2]
            [3]
            [0]
            [1]
            [2]
            [3]]]])
    """
    precision = None
    tol = 10 ** (-precision) if precision is not None else None

    if not isinstance(axis, int):
        raise TypeError(f"'{type(axis)}' cannot be interpreted as int")

    xs = [x.reformat(xs[0].orientation) for x in xs]
    axis = _to_positive_axis(axis, len(xs[0].shape), grow=False, invalid_axis=None)
    assert axis >= 0

    if axis in range(3):
        # Concatenate along spatial dimension
        for i, x in enumerate(xs[1:]):
            if not x._allclose_spacing(xs[0], precision=precision, ignore_origin=True):
                raise ValueError(
                    "All the inputs must have the same direction and pixel spacing "
                    "when concatenating spatial dimensions, but input at index 0 "
                    "has affine {} and the input at index {} "
                    "has affine {}".format(xs[0].affine[:3, :3], i, x.affine[:3, :3])
                )
        for i, (x1, x2) in enumerate(zip(xs[:-1], xs[1:])):
            ijk1 = np.array([0, 0, 0, 1])
            ijk1[axis] = x1.shape[axis]
            xyz = x1.affine.dot(ijk1)[:3]
            if not (
                (precision is not None and np.allclose(x2.scanner_origin, xyz, rtol=tol))
                or (np.asarray(x2.scanner_origin) == xyz).all()
            ):
                raise ValueError(
                    "All the inputs must be sequentially increasing in space "
                    "when concatenating spatial dimensions, but input at index {} "
                    "ends at xyz location {} and the input at index {} "
                    "starts at xyz location {}".format(i, xyz, i + 1, x2.scanner_origin)
                )
    else:
        for i, x in enumerate(xs[1:]):
            if not x._allclose_spacing(xs[0], precision=precision):
                raise ValueError(
                    "All the inputs must have the same affine matrix "
                    "when concatenating non-spatial dimensions, but input at index 0 "
                    "has affine {} and the input at index {} "
                    "has affine {}".format(xs[0].affine, i, x.affine)
                )

    volume = np.concatenate([x.volume for x in xs], axis=axis)
    headers = [x.headers() for x in xs]
    if any(x is None for x in headers):
        headers = None
    else:
        headers = np.concatenate(headers, axis=axis)
        if headers.ndim != volume.ndim or any(
            hs != 1 and hs != vs for hs, vs in zip(headers.shape, volume.shape)
        ):
            warnings.warn(
                "Got invalid headers shape ({}) given concatenated output shape ({}). "
                "Expected header dimensions to be 1 or same as volume dimension for all axes. "
                "Dropping all headers in concatenated output.".format(volume.shape, headers.shape)
            )
            headers = None

    return MedicalVolume(volume, xs[0].affine, headers=headers)


@implements(np.expand_dims)
def expand_dims(x, axis: Union[int, Sequence[int]]):
    """Expand across non-spatial dimensions.

    Args:
        x (MedicalVolume): A medical image.
        axis (``int(s)``): Axis/axes to expand dimensions.

    Returns:
        MedicalVolume: The medical image with expanded dimensions.

    Examples:
        >>> mv = vx.MedicalVolume(np.random.rand(3,4,5), affine=np.eye(4))
        >>> mv_expanded = np.expand_dims(mv, axis=-1)  # Expand last dimension
        >>> mv_expanded.shape
        (3, 4, 5, 1)
    """
    try:
        axis = _to_positive_axis(axis, len(x.shape), grow=True, invalid_axis="spatial")
    except ValueError:
        raise ValueError(f"Cannot expand across spatial dimensions (axis={axis})")
    vol = np.expand_dims(x.volume, axis)
    headers = x.headers()
    if headers is not None:
        headers = np.expand_dims(headers, axis)
    return x._partial_clone(volume=vol, headers=headers)


@implements(np.squeeze)
def squeeze(x, axis: Union[int, Sequence[int]] = None):
    """Squeeze non-spatial dimensions.

    Args:
        x (MedicalVolume): A medical image.
        axis (``int(s)``): Axis/axes to squeeze. Defaults to non-spatial axes.

    Returns:
        MedicalVolume: The medical image with squeezed dimensions.

    Raises:
        ValueError: If axis is not None, and an axis being squeezed is not of length 1
            or axis is not None and is squeezing spatial dimension (i.e. axis=0, 1, or 2).

    Examples:
        >>> mv = MedicalVolume(np.random.rand(3,4,5,1), np.eye(4))
        >>> mv_squeezed = np.squeeze(mv)  # squeeze all non-spatial dimensions
        >>> mv_squeezed.shape
        (3, 4, 5)
    """
    if axis is not None:
        try:
            axis = _to_positive_axis(axis, len(x.shape), grow=False, invalid_axis="spatial")
        except ValueError:
            raise ValueError(f"Cannot squeeze across spatial dimensions (axis={axis})")
    else:
        axis = tuple(i for i in range(3, len(x.shape)) if x.shape[i] == 1)
        if not axis:
            return x

    vol = np.squeeze(x.volume, axis=axis)
    headers = x.headers()
    if headers is not None:
        headers = np.squeeze(headers, axis=axis)

    return x._partial_clone(volume=vol, headers=headers)


@implements(np.pad)
def pad(x: MedicalVolume, pad_width, mode="constant", **kwargs):
    """Implementation of :func:`numpy.pad` for :class:`MedicalVolume`.

    Padding a MedicalVolume can affect the affine matrix of the volume.
    When spatial dimensions are padded, the scanner origin changes.

    In addition to standard numpy syntax for ``pad_width``, this
    function provides some shortcuts for padding particular dimensions.

    Either ``None`` or ``0`` can be used to indicate a dimension should
    not be padded. For example:

    >>> mv = MedicalVolume(np.ones(3,4,5), affine=np.eye(4))
    >>> pad(mv, (None, 0, (2,3)))  # dimensions 0 and 1 will not be padded
    >>> pad(mv, ((0,0), (0,0), (2,3)))  # equivalent to previous, but in numpy syntax

    Integers can also be used to indicate the dimension should be padded by the same
    amount on both sides:

    >>> mv = MedicalVolume(np.ones(3,4,5), affine=np.eye(4))
    >>> pad(mv, (5, (1,2), (2,3)))  # dimension 0 padded by 5 on both sides
    >>> pad(mv, ((5,5), (1,2), (2,3)))  # equivalent to previous, but in numpy syntax

    ``pad_width`` can also be shorter than the total MedicalVolume dimensions.
    In this case, the padding is applying in a broadcasting fashion. For example,
    if the MedicalVolume is 3D, then specifying padding widths for only two
    dimensions will pad the last two dimensions:

    >>> mv = MedicalVolume(np.ones(3,4,5), affine=np.eye(4))
    >>> pad(mv, (4, 6))  # last dimension padded by 6, second to last padded by 4
    >>> pad(mv, ((0,0), (4,4), (6,6)))  # equivalent to previous, but in numpy syntax

    Args:
        x (MedicalVolume): The medical image.
        pad_width (Union[Sequence, int]): Same as :func:`numpy.pad`.
        mode (str): Same as :func:`numpy.pad`.
        kwargs: Same as :func:`numpy.pad`.

    Returns:
        MedicalVolume: The padded medical image.

    Note:
        Currently, headers are not preserved upon padding. The returned medical image
        will not have any headers. This may change in the future.

    Examples:
        >>> arr = np.random.rand(3,4,5)
        >>> mv = MedicalVolume(arr, affine=np.eye(4))
        >>> mv_pad = np.pad(mv, 1)  # pad all dimensions by 1
    """
    if _is_int(pad_width):
        pad_width = ((pad_width,),) * x.ndim
    if len(pad_width) < x.ndim:
        pad_width = ((0,)) * (x.ndim - len(pad_width)) + tuple(pad_width)
    pad_width = tuple((0,) if x is None else (x,) if _is_int(x) else x for x in pad_width)
    pad_width = tuple(x * 2 if len(x) == 1 else x for x in pad_width)
    assert all(len(x) == 2 for x in pad_width), pad_width

    # Update scanner origin.
    ijk = np.asarray([-p[0] for p in pad_width[:3]] + [0])
    origin = x.affine @ ijk
    affine = x.affine.copy()
    affine[:, 3] = origin

    arr = np.pad(x.A, pad_width, mode=mode, **kwargs)

    return x._partial_clone(volume=arr, affine=affine, headers=None)


@implements(np.where)
def where(*args, **kwargs):
    """See :func:`numpy.where`."""
    return np.where(np.asarray(args[0]), *args[1:], **kwargs)


@implements(np.all)
def all_np(x, axis=None, keepdims=np._NoValue):
    """See :func:`numpy.all`."""
    return reduce_array_op(np.all, x, axis=axis, keepdims=keepdims)


@implements(np.any)
def any_np(x, axis=None, keepdims=np._NoValue):
    """See :func:`numpy.any`."""
    return reduce_array_op(np.any, x, axis=axis, keepdims=keepdims)


@implements(np.zeros_like)
def zeros_like(a, dtype=None, order="K", subok=True, shape=None):
    """See :func:`numpy.zeros_like`."""
    vol = np.zeros_like(a.A, dtype=dtype, order=order, subok=subok, shape=shape)
    return a._partial_clone(volume=vol)


@implements(np.ones_like)
def ones_like(a, dtype=None, order="K", subok=True, shape=None):
    """See :func:`numpy.ones_like`."""
    vol = np.ones_like(a.A, dtype=dtype, order=order, subok=subok, shape=shape)
    return a._partial_clone(volume=vol)


@implements(np.shares_memory)
def shares_memory(a, b, max_work=None):
    """Determine if two medical volumes share memory.

    This function implements :func:`numpy.shares_memory` for :class:`MedicalVolume`.
    Two volumes share memory if the pixel arrays and headers (if defined)
    share memory.

    Args:
        a (MedicalVolume): Input volume.
        b (MedicalVolume): Input volume.
        max_work (int, optional): Same as :func:`numpy.shares_memory`.

    Returns:
        bool: ``True`` if pixel arrays and headers (if defined) share memory.

    Raises:
        numpy.TooHardError: Exceeded max_work.

    Examples:
        >>> arr = np.random.rand(3,4,5)
        >>> mv1 = MedicalVolume(arr, affine=np.eye(4))
        >>> mv2 = MedicalVolume(arr, affine=np.eye(4))
        >>> np.shares_memory(mv1, mv2)  # Compare medicalVolume with same array in memory
        True
        >>> mv3 = MedicalVolume(arr.copy(), affine=np.eye(4))
        >>> np.shares_memory(mv1, mv2)  # Compare medicalVolume with different arrays in memory
        False
    """
    vol = np.shares_memory(a.A, b.A, max_work=max_work)
    headers = True
    if a.headers() is not None or b.headers() is not None:
        headers = np.shares_memory(a.headers(), b.headers(), max_work=max_work)
    return vol and headers


@implements(np.may_share_memory)
def may_share_memory(a, b, max_work=None):  # pragma: no cover
    """Determine if two medical volumes may share memory.

    This function implements :func:`numpy.may_share_memory` for :class:`MedicalVolume`.
    Two volumes share memory if the pixel arrays and headers (if defined)
    may share memory.

    Args:
        a (MedicalVolume): Input volume.
        b (MedicalVolume): Input volume.
        max_work (int, optional): Same as :func:`numpy.may_share_memory`.

    Returns:
        bool: ``True`` if pixel arrays and headers (if defined) may share memory.

    Raises:
        numpy.TooHardError: Exceeded max_work.

    Examples:
        >>> arr = np.random.rand(3,4,5)
        >>> mv1 = MedicalVolume(arr, affine=np.eye(4))
        >>> mv2 = MedicalVolume(arr, affine=np.eye(4))
        >>> np.shares_memory(mv1, mv2)  # Compare medicalVolume with same array in memory
        True
        >>> mv3 = MedicalVolume(arr.copy(), affine=np.eye(4))
        >>> np.shares_memory(mv1, mv2)  # Compare medicalVolume with different arrays in memory
        False
    """
    vol = np.may_share_memory(a.A, b.A, max_work=max_work)
    headers = True
    if a.headers() is not None or b.headers() is not None:
        headers = np.may_share_memory(a.headers(), b.headers(), max_work=max_work)
    return vol and headers


@implements(np.ascontiguousarray)
def ascontiguousarray(a, dtype=None, *, like=None):
    if isinstance(like, MedicalVolume):
        like = like.A
    if like is not None:
        raise ValueError("Does not support `like` argument.")
    return a._partial_clone(volume=np.ascontiguousarray(a.A, dtype=dtype))


def is_contiguous(x: MedicalVolume) -> bool:
    """Check if the pixel array is contiguous.

    If the underlying array (``mv.A``) is a CuPy array,
    returns ``True`` iff the array is C contiguous.

    Args:
        a (MedicalVolume): Input volume.

    Returns:
        bool: ``True`` if the pixel array is contiguous.
    """
    if x.device == cpu_device:
        # numpy
        return x.A.data.contiguous
    else:
        # cupy
        # TODO: check if it is sufficient to check c_contiguous
        return x.A.flags.c_contiguous


def _to_positive_axis(
    axis: Union[int, Sequence[int]],
    ndim: int,
    grow: bool = False,
    invalid_axis: Union[int, Sequence[int]] = None,
):
    """
    Args:
        axis (``int(s)``): The axis/axes to convert to positive.
        ndim (int): The current dimension.
        grow (bool, optional): If ``True``, converts axes to positive
            positions based on new dimension. The new dimension is
            calculated as ``ndim + #(axes < 0) + #(axes >= ndim)``.
        invalid_axis (bool, optional): Axes that are invalid.
            These should all be positive as check is done after
            positive.

    Returns:
        int(s): The positively formatted axes.
    """
    original_axis = axis

    is_sequence = isinstance(axis, Sequence)
    if not is_sequence:
        axis = (axis,)
    if grow:
        ndim += sum(tuple(x < 0 or x >= ndim for x in axis))
    axis = tuple(x if x >= 0 else ndim + x for x in axis)

    if invalid_axis is not None:
        if invalid_axis == "spatial":
            invalid_axis = tuple(range(0, 3))
        elif not isinstance(invalid_axis, Sequence):
            assert isinstance(invalid_axis, int)
            invalid_axis = (invalid_axis,)
        if any(x in invalid_axis for x in axis):
            raise ValueError(
                f"Invalid axes {original_axis}. Specified axes should not be in axes {invalid_axis}"
            )

    if not is_sequence:
        axis = axis[0]
    return axis


def _is_int(x):
    return isinstance(x, int) or (
        np.isscalar(x) and hasattr(x, "dtype") and np.issubdtype(x.dtype, np.integer)
    )



from typing import Tuple

import numpy as np
import pydicom

from voxel.device import get_array_module

__all__ = [
    "apply_rescale",
    "apply_window",
    "pixel_dtype",
    "pixel_range",
    "invert",
    "invert_window",
]


def pixel_dtype(ds: "pydicom.Dataset") -> np.dtype:
    """Detect the NumPy dtype of the data in a header.
    NB: byteorder can be incorrect, so use with caution (see pydicom pixel_dtype docs).

    Args:
        ds (pydicom.Dataset): DICOM header.

    Returns:
        np.dtype: NumPy dtype of the data.
    """

    if "FloatPixelData" in ds:
        return np.float32

    elif "DoubleFloatPixelData" in ds:
        return np.float64

    # Default to PixelData
    elif "BitsAllocated" in ds:
        bits_allocated = ds.BitsAllocated
        if bits_allocated in [8, 16, 32, 64]:
            prefix = "int" if ds.get("PixelRepresentation", 0) == 1 else "uint"
            return np.dtype(f"{prefix}{bits_allocated}")

        raise ValueError(f"Unsupported bit depth {bits_allocated}")

    raise ValueError("NumPy dtype could not be detected from the headers.")


def pixel_range(ds: "pydicom.Dataset") -> Tuple[float, float]:
    """Get the pixel range of a DICOM dataset.

    Args:
        ds (pydicom.Dataset): DICOM header.

    Returns:
        Tuple[float, float]: Pixel range.
    """
    if "FloatPixelData" in ds or "DoubleFloatPixelData" in ds:
        raise ValueError("Pixel range cannot be determined for float pixel data.")

    if "BitsStored" in ds or "BitsAllocated" in ds:
        bits = ds.get("BitsStored", ds.get("BitsAllocated"))
        pixel_representation = ds.get("PixelRepresentation", 0)
        if pixel_representation == 1:
            return -(2 ** (bits - 1)), 2 ** (bits - 1) - 1

        return 0, 2**bits - 1

    raise ValueError("Pixel range could not be detected from the headers.")


def apply_window(
    volume: np.ndarray,
    center: float,
    width: float,
    output_range: Tuple[float, float] = None,
    mode: str = None,
    inplace: bool = False,
) -> np.ndarray:
    """Apply a window to a volume.

    The output range defaults to the boundaries of the window in contrast to
    the default behaviour of pydicom which is to use the pixel range. Explicitly
    setting the output range to the pixel range will result in the same behaviour.
    Set the output range to (0, 1) to normalize the volume.

    NB: Whether the minimum output value is rendered as black or white may depend
    on the value of Photometric Interpretation (0028,0004).

    Args:
        volume (np.ndarray): Volume to apply the window to.
        center (float): Window center.
        width (float): Window width.
        output_range (Tuple[float, float]): Output range.
        mode (str, optional): VOI LUT function. Defaults to None.
        inplace (bool, optional): Whether to apply the window in place. Defaults to False.

    Returns:
        np.ndarray: Windowed volume.
    """
    mode = mode.lower() if mode is not None else "linear"

    wc, ww = center, width
    if mode == "linear":
        if ww < 1:
            raise ValueError("Window Width must be greater than 1 for LINEAR.")
        wc -= 0.5
        ww -= 1

    if mode == "linear_exact" and ww <= 0:
        raise ValueError("Window Width must be greater than 0 for LINEAR_EXACT.")

    lb, ub = wc - ww / 2, wc + ww / 2
    if output_range is None:
        output_range = (lb, ub)

    y_min, y_max = output_range
    y_range = y_max - y_min

    volume = volume if inplace else volume.copy()
    xp = get_array_module(volume)
    if mode in ["linear", "linear_exact"]:
        volume -= wc
        volume /= ww
        volume += 0.5
        xp.clip(volume, 0, 1, out=volume)

        if y_range != 1:
            volume *= y_range

        if y_min != 0:
            volume += y_min

    elif mode == "sigmoid":
        if ww <= 0:
            raise ValueError("Window Width must be greater than 0 for SIGMOID.")

        # volume[:] = y_range / (1 + xp.exp(-4 * (volume - wc) / ww)) + y_min
        volume -= wc
        volume *= -4 / wc
        xp.exp(volume, out=volume)
        volume += 1
        xp.divide(y_range, volume, out=volume)
        volume += y_min

    else:
        raise ValueError(f"VOI LUT Function {mode} is not supported.")

    return volume


def apply_rescale(
    volume: np.ndarray,
    slope: float,
    intercept: float,
    inplace: bool = False,
) -> np.ndarray:
    """Apply a rescale to a volume.

    Args:
        volume (np.ndarray): Volume to apply the rescale to.
        slope (float): Rescale slope.
        intercept (float): Rescale intercept.
        inplace (bool, optional): Whether to apply the rescale in place. Defaults to False.

    Returns:
        np.ndarray: Rescaled volume.
    """
    volume = volume if inplace else volume.copy()
    if slope != 1:
        volume *= slope
    if intercept != 0:
        volume += intercept
    return volume


def invert(
    volume: np.ndarray,
    output_range: Tuple[float, float] = None,
    inplace: bool = False,
) -> np.ndarray:
    """Invert the pixel values in a volume.

    Args:
        volume (np.ndarray): Volume to invert.
        pixel_range (Tuple[float, float], optional): Pixel range. Defaults to None.
        inplace (bool, optional): Whether to apply the rescale in place. Defaults to False.

    Returns:
        np.ndarray: Inverted volume.
    """
    if not inplace:
        volume = volume.copy()

    x_min, x_max = volume.min(), volume.max()
    x_range = x_max - x_min

    volume -= x_min
    volume /= x_range

    xp = get_array_module(volume)
    xp.subtract(1, volume, out=volume)

    # Rescale to output range
    if output_range is None:
        output_range = (x_min, x_max)

    if output_range != (0, 1):
        lb, ub = output_range
        volume *= ub - lb
        volume += lb

    return volume


def invert_window(
    volume: np.ndarray,
    center: float,
    width: float,
    output_range: Tuple[float, float] = None,
) -> Tuple[float, float]:
    """Invert the window parameters.

    Args:
        volume (np.ndarray): Volume to invert the window for.
        center (float): Window center.
        width (float): Window width.
        output_range (Tuple[float, float], optional): Output range. Defaults to None.

    Returns:
        Tuple[float, float]: Inverted window parameters.
    """
    x_min, x_max = volume.min(), volume.max()
    x_range = x_max - x_min

    # Rescale the window
    center = 1 - ((center - x_min) / x_range)
    width = width / x_range

    # Rescale to output range
    if output_range is None:
        output_range = (x_min, x_max)

    if output_range != (0, 1):
        lb, ub = output_range
        center = center * (ub - lb) + lb
        width = width * (ub - lb)

    return center, width



import unittest

import numpy as np

from voxel.device import Device, cpu_device, get_device, to_device
from voxel.med_volume import MedicalVolume

from .util import requires_packages


class TestDevice(unittest.TestCase):
    def test_basic(self):
        assert Device(-1) == cpu_device
        assert Device("cpu") == cpu_device
        assert cpu_device.xp == np

        device = Device(-1)
        assert int(device) == -1
        assert device.index == -1
        assert device.id == -1
        assert device == -1
        assert device.cpdevice is None

        device2 = Device(-1)
        assert device2 == device

    @requires_packages("cupy")
    def test_cupy(self):
        import cupy as cp

        device = Device(0)
        assert device.cpdevice == cp.cuda.Device(0)
        assert device.type == "cuda"
        assert device.index == 0
        assert device.xp == cp
        assert int(device) == 0

        device = Device(cp.cuda.Device(0))
        assert device.cpdevice == cp.cuda.Device(0)
        assert device.type == "cuda"
        assert device.index == 0

    @requires_packages("sigpy")
    def test_sigpy(self):
        import sigpy as sp

        assert Device(-1) == sp.cpu_device
        assert Device(sp.cpu_device) == sp.cpu_device

        device = Device(-1)
        assert device == sp.cpu_device
        assert device.spdevice == sp.cpu_device

    @requires_packages("sigpy", "cupy")
    def test_sigpy_cupy(self):
        import sigpy as sp

        assert Device(0) == sp.Device(0)

        device = Device(0)
        assert device.spdevice == sp.Device(0)

    @requires_packages("torch")
    def test_torch(self):
        import torch

        pt_device = torch.device("cpu")

        assert Device(pt_device) == cpu_device

        vx_device = Device(-1)
        assert vx_device == pt_device
        assert vx_device.ptdevice == pt_device

    def test_to_device(self):
        arr = np.ones((3, 3, 3))
        mv = MedicalVolume(arr, affine=np.eye(4))

        arr2 = to_device(arr, -1)
        assert get_device(arr2) == cpu_device

        mv2 = to_device(mv, -1)
        assert get_device(mv2) == cpu_device


if __name__ == "__main__":
    unittest.main()



"""Standardized orientation convention and utilities.

Medical image orientation convention is library and image format (DICOM, NIfTI, etc.)
dependent and is often difficult to interpret. This makes it challenging to intelligently
and rapidly reformat images.

We adopt a human readable orientation representation
for the dimensions and define utilities to convert between different orientation formats
from current libraries (Nibabel, PyDicom, ITK, etc).

Orientations are represented by string axis codes:

- ``"LR"``: left to right; ``"RL"``: right to left
- ``"PA"``: posterior to anterior; ``"AP"``: anterior to posterior
- ``"IS"``: inferior to superior; ``"SI"``: superior to inferior

A :class:`MedicalVolume` object with orientation ``("SI", "AP", "LR")`` has an
array where the first dimension spans superior -> inferior, the second dimension
spans anterior -> posterior, and the third dimension spans left -> right. Voxel
at (i,j,k) index ``(0,0,0)`` would be the (superior, anterior, left) corner.

In many cases, images are not acquired in the standard plane convention, but rather
in a rotated frame. In this case, the orientations correspond to the closest axis
the a particular dimension.

Two general conventions are followed:

- All orientations are in patient voxel coordinates. Image data from (i, j, k)
  corresponds to the voxel at array position ``arr[i,j,k]``.
- Left: corresponds to patient (not observer) left,
  right: corresponds to patient (not observer) right.

We adopt the RAS+ standard (as defined by NIfTI) for orienting our images.
The ``+`` in RAS+ indicates that all directions point to the increasing direction.
i.e. from -x to x:.

Image spacing, direction, and global origin are represented by a 4x4 affine matrix (:math:`A`) and
is identical to the nibabel affine matrix
(see `nibabel <https://nipy.org/nibabel/coordinate_systems.html>`_).
The affine matrix converts pixel coordinates (i, j, k) into world (NIfTI) coordinates (x, y, z).

.. math::

    \\begin{bmatrix} x\\\\y\\\\z\\\\1\\end{bmatrix} = A
    \\begin{bmatrix} i\\\\j\\\\k\\\\1\\end{bmatrix}


For example,

.. math::

    \\begin{bmatrix} x\\\\y\\\\z\\\\1 \\end{bmatrix} =
    \\begin{bmatrix} 0 & 0 & 1.5 & -61.6697\\\\-0.3125 & 0 & 0 & 50.8516\\\\
    0 & -0.3125 & 0 & 88.5876\\\\0 & 0 & 0 & 1 \\end{bmatrix}
    \\begin{bmatrix} i\\\\j\\\\k\\\\1\\end{bmatrix}

For details on how the affine matrix is used for reformatting see
:class:`voxel.MedicalVolume`.
"""
import warnings
from typing import List, Sequence, Tuple, Union

import nibabel.orientations as nibo
import numpy as np
import pydicom

import voxel as vx

__all__ = [
    "to_affine",
    "to_RAS_affine",
    "get_transpose_inds",
    "get_flip_inds",
    "orientation_nib_to_standard",
    "orientation_standard_to_nib",
    "SAGITTAL",
    "CORONAL",
    "AXIAL",
]


SAGITTAL = ("SI", "AP", "LR")
CORONAL = ("SI", "LR", "AP")
AXIAL = ("AP", "LR", "SI")

__EXPECTED_ORIENTATION_TUPLE_LEN__ = 3
__SUPPORTED_ORIENTATIONS__ = ["LR", "RL", "PA", "AP", "IS", "SI"]
__ORIENTATIONS_TO_AXIS_ID__ = {"LR": 0, "RL": 0, "PA": 1, "AP": 1, "IS": 2, "SI": 2}


def __check_orientation__(orientation: tuple):
    """Check if orientation tuple defines a valid orientation.

    Args:
        orientation (tuple[str]): Image orientation in standard orientation format.

    Raises:
        ValueError: If orientation tuple is invalid.
    """
    is_orientation_format = (
        len(orientation) == __EXPECTED_ORIENTATION_TUPLE_LEN__
        and sum([type(o) is str for o in orientation]) == __EXPECTED_ORIENTATION_TUPLE_LEN__
    )

    orientation_str_exists = (
        sum([o in __SUPPORTED_ORIENTATIONS__ for o in orientation])
        == __EXPECTED_ORIENTATION_TUPLE_LEN__
    )

    orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in orientation]
    unique_ids = len(orientation_ids) == len(set(orientation_ids))

    if not is_orientation_format or not orientation_str_exists or not unique_ids:
        raise ValueError(
            "Orientation format mismatch: Orientations must be tuple of strings of "
            "length {}".format(__EXPECTED_ORIENTATION_TUPLE_LEN__)
        )


def get_transpose_inds(curr_orientation: tuple, new_orientation: tuple):
    """Get indices for reordering planes from ``curr_orientation`` to ``new_orientation``.

    Only permuted order of reformatting the image planes is returned.
    For example, ``("SI", "AP", "LR")`` and ``("IS", "PA", "RL")`` will have no permuted
    indices because "SI"/"IS", "AP"/"PA" and "RL"/"LR" each correspond to the same
    plane.

    Args:
        curr_orientation (tuple[str]): Current image orientation.
        new_orientation (tuple[str]): New image orientation.

    Returns:
        tuple[int]: Axes to transpose to change orientation.

    Examples:
        >>> get_transpose_inds(("SI", "AP", "LR"), ("AP", "SI", "LR"))
        (1,0,2)
        >>> get_transpose_inds(("SI", "AP", "LR"), ("IS", "PA", "RL"))
        (0,1,2)
    """
    __check_orientation__(curr_orientation)
    __check_orientation__(new_orientation)

    curr_orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in curr_orientation]
    new_orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in new_orientation]

    if set(curr_orientation_ids) != set(new_orientation_ids):
        raise ValueError(
            "Orientation mismatch: Both curr_orientation and new_orientation "
            "must contain the same axes"
        )

    transpose_inds = [curr_orientation_ids.index(n_o) for n_o in new_orientation_ids]

    return tuple(transpose_inds)


def get_flip_inds(curr_orientation: tuple, new_orientation: tuple):
    """Get indices to flip from ``curr_orientation`` to ``new_orientation``.

    Args:
        curr_orientation (tuple[str]): Current image orientation.
        new_orientation (tuple[str]): New image orientation.

    Returns:
        list[int]: Axes to flip.

    Raises:
        ValueError: If mismatch in orientation indices. To avoid this error,
            use :func:`get_transpose_inds` prior to flipping.

    Examples:
        >>> get_transpose_inds(("SI", "AP", "LR"), ("IS", "AP", "RL"))
        [0, 2]
    """
    __check_orientation__(curr_orientation)
    __check_orientation__(new_orientation)

    curr_orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in curr_orientation]
    new_orientation_ids = [__ORIENTATIONS_TO_AXIS_ID__[o] for o in new_orientation]

    if curr_orientation_ids != new_orientation_ids:
        raise ValueError(
            "All axis orientations (S/I, L/R, A/P) must be ordered. "
            "Use `get_transpose_inds` to reorder axes."
        )

    flip_axs_inds = []
    for i in range(__EXPECTED_ORIENTATION_TUPLE_LEN__):
        if curr_orientation[i] != new_orientation[i]:
            flip_axs_inds.append(i)

    return flip_axs_inds


# Nibabel to standard orientation conversion utils.
__nib_to_standard_orientation_map__ = {
    "R": "LR",
    "L": "RL",
    "A": "PA",
    "P": "AP",
    "S": "IS",
    "I": "SI",
}


def orientation_nib_to_standard(nib_orientation):
    """Convert Nibabel orientation to the standard voxel orientation format.

    Args:
        nib_orientation: a RAS+ tuple orientation used by Nibabel.

    Returns:
        tuple[str]: Image orientation in the standard orientation format.

    Examples:
        >>> orientation_nib_to_standard(("R", "A", "S"))
        ("LR", "PA", "IS")
    """
    orientation = []
    for symb in nib_orientation:
        orientation.append(__nib_to_standard_orientation_map__[symb])
    return tuple(orientation)


def orientation_standard_to_nib(orientation):
    """Convert standard voxel orientation format to Nibabel orientation.

    Args:
        orientation: Image orientation in the standard orientation format.

    Returns:
        tuple[str]: RAS+ tuple orientation used by Nibabel.

    Examples:
        >>> orientation_nib_to_standard(("LR", "PA", "IS"))
        ("R", "A", "S")
    """
    nib_orientation = []
    for symb in orientation:
        nib_orientation.append(symb[1])
    return tuple(nib_orientation)


def to_affine(
    orientation,
    spacing: Sequence[Union[int, float]] = None,
    origin: Sequence[Union[int, float]] = None,
):
    """Convert orientation, spacing, and origin data into affine matrix.

    Args:
        orientation (Sequence[str]): Image orientation in the standard orientation format
            (e.g. ``("LR", "AP", "SI")``).
        spacing (int(s) | float(s)): Number(s) corresponding to pixel spacing of each direction.
            If a single value, same pixel spacing is used for all directions.
            If sequence is less than length of ``orientation``, remaining direction have unit
            spacing (i.e. ``1``). Defaults to unit spacing ``(1, 1, 1)``
        origin (int(s) | float(s)): The ``(x0, y0, z0)`` origin for the scan.
            If a single value, same origin is used for all directions.
            If sequence is less than length of ``orientation``, remaining direction have standard
            origin (i.e. ``0``). Defaults to ``(0, 0, 0)``

    Returns:
        ndarray: A 4x4 ndarray representing the affine matrix.

    Examples:
        >>> to_affine(("SI", "AP", "RL"), spacing=(0.5, 0.5, 1.5), origin=(10, 20, 0))
        array([[-0. , -0. , -1.5,  10. ],
               [-0. , -0.5, -0. ,  20. ],
               [-0.5, -0. , -0. ,  30. ],
               [ 0. ,  0. ,  0. ,   1. ]])

    Note:
        This method assumes all direction follow the standard principal directions in the normative
        patient orientation. Moving along one direction of the array only moves along one fo the
        normative directions.
    """

    def _format_numbers(input, default_val, name, expected_num):
        """Formats (sequence of) numbers (spacing, origin) into standard 3-length tuple."""
        if input is None:
            return (default_val,) * expected_num
        if isinstance(input, (int, float)):
            return (input,) * expected_num

        if not isinstance(input, (np.ndarray, Sequence)) or len(input) > expected_num:
            raise ValueError(
                f"`{name}` must be a real number or sequence (length<={expected_num}) "
                f"of real numbers. Got {input}"
            )
        input = tuple(input)

        if len(input) < expected_num:
            input += (default_val,) * (expected_num - len(input))
        assert len(input) == expected_num

        return input

    if len(orientation) == 2:
        orientation = _infer_orientation(orientation)
    __check_orientation__(orientation)
    spacing = _format_numbers(spacing, 1, "spacing", len(orientation))
    origin = _format_numbers(origin, 0, "origin", len(orientation))

    affine = np.eye(4)
    start_ornt = nibo.io_orientation(affine)
    end_ornt = nibo.axcodes2ornt(orientation_standard_to_nib(orientation))
    ornt = nibo.ornt_transform(start_ornt, end_ornt)

    transpose_idxs = ornt[:, 0].astype(int)
    flip_idxs = ornt[:, 1]

    affine[:3] = affine[:3][transpose_idxs]
    affine[:3] *= flip_idxs[..., np.newaxis]
    affine[:3, :3] *= np.asarray(spacing)
    affine[:3, 3] = np.asarray(origin)

    return affine


def _infer_orientation(orientation):
    """Infer 3-length orientation from 2-length orientation.

    Args:
        orientation: The incomplete orientation.

    Returns:
        tuple[str, str, str]: Standard orientation.
    """
    idxs = {__ORIENTATIONS_TO_AXIS_ID__[k] for k in orientation}
    if len(orientation) != 2 or len(idxs) != 2:
        raise ValueError(
            "`orientation` must be an incomplete orientation that encodes orthogonal directions"
        )

    missing_ornt = [k for k, v in __ORIENTATIONS_TO_AXIS_ID__.items() if v not in idxs][0]
    return tuple(orientation) + (missing_ornt,)


def to_RAS_affine(headers: List[pydicom.FileDataset], default_ornt: Tuple[str, str, str] = None):
    """Convert from LPS+ orientation (default for DICOM) to RAS+ standardized orientation.

    Args:
        headers (list[pydicom.FileDataset]): Headers for DICOM files to reorient.
            Files should correspond to single volume.

    Returns:
        np.ndarray: Affine matrix.
    """

    # The location of the DICOM attributes relevant to determine the affine
    # matrix are different depending on whether the DICOM file is enhanced or
    # in legacy format. Recall that most of contemporary DICOM files are still
    # (CT > Enhanced CT), but some are always enhanced, such as DBT.
    h0 = headers[0]
    if "PerFrameFunctionalGroupsSequence" not in h0:
        iop = h0.get("ImageOrientationPatient", None)
        if iop is None:
            iop = _decode_inplane_direction(headers, default_ornt=default_ornt)

        in_plane_pixel_spacing = h0.get("PixelSpacing", None)
        if in_plane_pixel_spacing is None:
            in_plane_pixel_spacing = h0.get("ImagerPixelSpacing", None)

        ipp0 = h0.get("ImagePositionPatient", None)
        ipp1 = headers[1].get("ImagePositionPatient", None) if len(headers) > 1 else None
        slice_thickness = h0.get("SliceThickness", None)
        slice_spacing = h0.get("SpacingBetweenSlices", None)
    else:
        iop = _get_enhanced_attr(h0, ["PlaneOrientationSequence", "ImageOrientationPatient"])
        in_plane_pixel_spacing = _get_enhanced_attr(h0, ["PixelMeasuresSequence", "PixelSpacing"])
        ipp0 = _get_enhanced_attr(h0, ["PlanePositionSequence", "ImagePositionPatient"], idx=0)
        ipp1 = _get_enhanced_attr(h0, ["PlanePositionSequence", "ImagePositionPatient"], idx=1)
        slice_thickness = _get_enhanced_attr(h0, ["PixelMeasuresSequence", "SliceThickness"]) / len(
            h0.PerFrameFunctionalGroupsSequence
        )
        slice_spacing = _get_enhanced_attr(h0, ["PixelMeasuresSequence", "SpacingBetweenSlices"])

    # Fallbacks
    if iop is None:
        iop = [1, 0, 0, 0, 1, 0]
        warnings.warn("Could not determine direction cosines, using LPS+ as a fallback.")

    if in_plane_pixel_spacing is None:
        in_plane_pixel_spacing = [1, 1]
        warnings.warn("Could not determine in-plane pixel spacing, using [1, 1] as a fallback.")

    if ipp0 is None:
        ipp0 = [0, 0, 0]
        warnings.warn("Could not determine image position, using [0, 0, 0] as a fallback.")

    orientation = np.zeros([3, 3])

    # Determine vector for in-plane pixel directions (i, j).
    i_vec, j_vec = np.asarray(iop[:3]).astype(np.float64), np.asarray(iop[3:]).astype(np.float64)
    i_vec, j_vec = (
        np.round(i_vec, vx.config.affine_precision),
        np.round(j_vec, vx.config.affine_precision),
    )
    i_vec = i_vec * in_plane_pixel_spacing[0]
    j_vec = j_vec * in_plane_pixel_spacing[1]

    # Determine vector for through-plane pixel direction (k).
    # Compute difference in patient position between consecutive headers.
    # This is the preferred method to determine the k vector.
    # If single header, take cross product between i/j vectors.
    # These actions are done to avoid rounding errors that might result from float subtraction.
    if ipp1 is not None and ipp1 != ipp0:
        k_vec = np.asarray(ipp1).astype(np.float64) - np.asarray(ipp0).astype(np.float64)
    else:
        if slice_thickness is None:
            slice_thickness = 1.0
            warnings.warn("Could not determine slice thickness, using 1.0 as a fallback.")

        i_norm = 1 / np.linalg.norm(i_vec) * i_vec
        j_norm = 1 / np.linalg.norm(j_vec) * j_vec
        k_norm = np.cross(i_norm, j_norm)
        k_vec = k_norm / np.linalg.norm(k_norm) * slice_thickness
        if isinstance(slice_spacing, (int, float)) and slice_spacing < 0:
            k_vec *= -1

    k_vec = np.round(k_vec, vx.config.affine_precision)

    orientation[:3, :3] = np.stack([j_vec, i_vec, k_vec], axis=1)
    scanner_origin = np.asarray(ipp0).astype(np.float64)
    scanner_origin = np.round(scanner_origin, vx.config.affine_precision)

    affine = np.zeros([4, 4])
    affine[:3, :3] = orientation
    affine[:3, 3] = scanner_origin
    affine[:2, :] = -1 * affine[:2, :]
    affine[3, 3] = 1

    affine[affine == 0] = 0

    return affine


def _decode_inplane_direction(headers: Sequence[pydicom.FileDataset], default_ornt=None):
    """Helper function to decode in-plane direction from header(s).

    Recall the direction in dicoms are in cartesian order ``(x,y)``,
    but numpy/voxel are in matrix order ``(y,x)``. When adding new
    methods, make sure to account for this.

    Returns:
        np.ndarray: 6-element LPS direction array where first 3 elements define
            direction for x-direction (columns) and second 3 elements define
            direction for y-direction (rows)
    """
    _patient_ornt_to_nib = {"H": "S", "F": "I"}

    if (
        len(headers) == 1
        and hasattr(headers[0], "PatientOrientation")
        and headers[0].PatientOrientation
    ):
        # Decoder: patient orientation.
        # Patient orientation is only decoded along principal direction (e.g. "FR" -> "F").
        ornt = [
            _patient_ornt_to_nib.get(k[:1], k[:1]) for k in headers[0].PatientOrientation
        ]  # (x,y)
        ornt = orientation_nib_to_standard(ornt)
        affine = to_affine(ornt)
        affine[:2, :] = -1 * affine[:2, :]
        return np.concatenate([affine[:3, 0], affine[:3, 1]], axis=0)

    if default_ornt:
        affine = to_affine(default_ornt)
        affine[:2, :] = -1 * affine[:2, :]
        return np.concatenate([affine[:3, 0], affine[:3, 1]], axis=0)

    return None


def _unroll_dicom_attr(header, attr: List[pydicom.tag.TagType], idx: int = 0):
    for a in attr[:-1]:
        if a not in header:
            return None

        if isinstance(header[a], pydicom.sequence.Sequence):
            if a == pydicom.tag.Tag("PerFrameFunctionalGroupsSequence"):
                header = header[a][idx]
            else:
                header = header[a][0]

        return _unroll_dicom_attr(header[a][0], attr[1:])

    value = header.get(attr[-1], None)
    if isinstance(value, pydicom.DataElement):
        value = value.value

    return value


def _get_enhanced_attr(header, attr: List[pydicom.tag.TagType], idx: int = 0):
    """Get attribute value from enhanced dicom header.

    Returns:
        Any: Attribute value.
    """

    frame = _unroll_dicom_attr(header, ["PerFrameFunctionalGroupsSequence", *attr], idx)
    if frame is not None:
        return frame

    return _unroll_dicom_attr(header, ["SharedFunctionalGroupsSequence", *attr], idx)



"""Tests for numpy routines and ufunc on MedicalVolumes."""
import unittest

import numpy as np

from voxel.med_volume import MedicalVolume

from . import util as ututils


class TestNumpyRoutinesMedicalVolume(unittest.TestCase):
    _AFFINE = np.asarray(
        [
            [0.0, 0.0, 0.8, -171.41],
            [0.0, -0.3125, 0.0, 96.0154],
            [-0.3125, 0.0, 0.0, 47.0233],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )  # ('SI', 'AP', 'LR')

    def _build_volume(self, shape_or_vol=(10, 20, 30, 2), affine=None):
        if isinstance(shape_or_vol, np.ndarray):
            vol = shape_or_vol
            shape = shape_or_vol.shape
        elif isinstance(shape_or_vol, MedicalVolume):
            vol = shape_or_vol.A
            shape = shape_or_vol.shape
        else:
            shape = shape_or_vol
            vol = np.random.rand(*shape)

        if len(shape) != 4:
            raise ValueError(f"shape must correspond to 4D volume. Got {shape}")

        headers = np.stack(
            [
                ututils.build_dummy_headers(shape[2], {"EchoTime": 2}),
                ututils.build_dummy_headers(shape[2], {"EchoTime": 10}),
            ],
            axis=-1,
        )
        if affine is None:
            affine = np.eye(4)
        return MedicalVolume(vol, affine, headers=headers)

    def test_where(self):
        mv = MedicalVolume(np.ones((10, 20, 30)), np.eye(4))
        mv[np.where(mv == 1)] = 5
        assert np.all(mv == 5)
        assert not np.any(mv == 1)
        assert all(mv == 5)
        assert not any(mv == 5)

    def test_exp(self):
        mv = MedicalVolume(np.ones((10, 20, 30)), np.eye(4))
        assert np.all(np.exp(mv.volume) == np.exp(mv).volume)

    def test_ndarray_arithmetic(self):
        mv = MedicalVolume(np.ones((10, 20, 30)), np.eye(4))
        mv2 = mv + np.ones(mv.shape)
        assert all(mv2 == 2)

    def test_reduce_funcs(self):
        shape = (10, 20, 30, 2)
        mv = self._build_volume(shape)

        mv2 = np.add.reduce(mv, -1)
        assert np.all(mv2 == np.add.reduce(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.add.reduce(mv, axis=None)
        assert np.all(mv2 == np.add.reduce(mv.volume, axis=None))
        assert np.isscalar(mv2)

    def test_stats_funcs(self):
        shape = (10, 20, 30, 2)
        mv = self._build_volume(shape)

        mv2 = np.sum(mv, axis=-1)
        assert np.all(mv2 == np.sum(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.sum(mv)
        assert np.all(mv2 == np.sum(mv.volume))
        assert np.isscalar(mv2)

        mv2 = mv.sum(axis=-1)
        assert np.all(mv2 == np.sum(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = mv.sum()
        assert np.all(mv2 == np.sum(mv.volume))
        assert np.isscalar(mv2)

        mv2 = np.mean(mv, axis=-1)
        assert np.all(mv2 == np.mean(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.mean(mv)
        assert np.all(mv2 == np.mean(mv.volume))
        assert np.isscalar(mv2)

        mv2 = mv.mean(axis=-1)
        assert np.all(mv2 == np.mean(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = mv.mean()
        assert np.all(mv2 == np.mean(mv.volume))
        assert np.isscalar(mv2)

        mv2 = np.std(mv, axis=-1)
        assert np.all(mv2 == np.std(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.std(mv)
        assert np.all(mv2 == np.std(mv.volume))
        assert np.isscalar(mv2)

        # Min/max functions
        mv2 = np.amin(mv, axis=-1)
        assert np.all(mv2 == np.amin(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.amin(mv)
        assert np.all(mv2 == np.amin(mv.volume))
        assert np.isscalar(mv2)

        mv2 = np.amax(mv, axis=-1)
        assert np.all(mv2 == np.amax(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.amax(mv)
        assert np.all(mv2 == np.amax(mv.volume))
        assert np.isscalar(mv2)

        mv2 = np.argmin(mv, axis=-1)
        assert np.all(mv2 == np.argmin(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.argmin(mv)
        assert np.all(mv2 == np.argmin(mv.volume))

        mv2 = np.argmax(mv, axis=-1)
        assert np.all(mv2 == np.argmax(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.argmax(mv)
        assert np.all(mv2 == np.argmax(mv.volume))

    def test_nan_funcs(self):
        shape = (10, 20, 30, 2)
        mv = self._build_volume(shape)
        headers = mv.headers

        vol_nan = np.ones(shape)
        vol_nan[..., 1] = np.nan
        mv = MedicalVolume(vol_nan, np.eye(4), headers=headers)

        mv2 = np.nansum(mv, axis=-1)
        assert np.all(mv2 == np.nansum(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.nansum(mv)
        assert np.all(mv2 == np.nansum(mv.volume))
        assert np.isscalar(mv2)

        mv2 = np.nanmean(mv, axis=-1)
        assert np.all(mv2 == np.nanmean(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.nanmean(mv)
        assert np.all(mv2 == np.nanmean(mv.volume))
        assert np.isscalar(mv2)

        mv2 = np.nanstd(mv, axis=-1)
        assert np.all(mv2 == np.nanstd(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.nanstd(mv)
        assert np.all(mv2 == np.nanstd(mv.volume))
        assert np.isscalar(mv2)

        mv2 = np.nan_to_num(mv)
        assert np.unique(mv2.volume).tolist() == [0, 1]
        mv2 = np.nan_to_num(mv, copy=False)
        assert id(mv2) == id(mv)

        mv2 = np.nanmin(mv, axis=-1)
        assert np.all(mv2 == np.nanmin(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.nanmin(mv)
        assert np.all(mv2 == np.nanmin(mv.volume))
        assert np.isscalar(mv2)

        mv2 = np.nanmax(mv, axis=-1)
        assert np.all(mv2 == np.nanmax(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.nanmax(mv)
        assert np.all(mv2 == np.nanmax(mv.volume))
        assert np.isscalar(mv2)

        mv2 = np.nanargmin(mv, axis=-1)
        assert np.all(mv2 == np.nanargmin(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.nanargmin(mv)
        assert np.all(mv2 == np.nanargmin(mv.volume))

        mv2 = np.nanargmax(mv, axis=-1)
        assert np.all(mv2 == np.nanargmax(mv.volume, axis=-1))
        assert mv2.shape == mv.shape[:3]
        mv2 = np.nanargmax(mv)
        assert np.all(mv2 == np.nanargmax(mv.volume))

    def test_round(self):
        shape = (10, 20, 30, 2)
        affine = np.concatenate([np.random.rand(3, 4), [[0, 0, 0, 1]]], axis=0)
        mv = self._build_volume(shape, affine=affine)

        mv2 = mv.round()
        assert np.allclose(mv2.affine, affine)
        assert np.unique(mv2.volume).tolist() == [0, 1]

        mv2 = mv.round(affine=True)
        assert np.unique(mv2.affine).tolist() == [0, 1]
        assert np.unique(mv2.volume).tolist() == [0, 1]

    def test_clip(self):
        # Clip
        shape = (10, 20, 30)
        mv = MedicalVolume(np.random.rand(*shape), np.eye(4))

        mv2 = np.clip(mv, 0.4, 0.6)
        assert np.all((mv2.volume >= 0.4) & (mv2.volume <= 0.6))

        mv_lower = MedicalVolume(np.ones(mv.shape) * 0.4, mv.affine)
        mv_upper = MedicalVolume(np.ones(mv.shape) * 0.6, mv.affine)
        mv2 = np.clip(mv, mv_lower, mv_upper)
        assert np.all((mv2.volume >= 0.4) & (mv2.volume <= 0.6))

    def test_array_like(self):
        shape = (10, 20, 30)
        mv = MedicalVolume(np.random.rand(*shape), np.eye(4))

        mv2 = np.zeros_like(mv)
        assert np.all(mv2.volume == 0)

        mv2 = np.ones_like(mv)
        assert np.all(mv2.volume == 1)

    def test_shares_memory(self):
        shape = (10, 20, 30, 2)
        headers = np.stack(
            [
                ututils.build_dummy_headers(shape[2], {"EchoTime": 2}),
                ututils.build_dummy_headers(shape[2], {"EchoTime": 10}),
            ],
            axis=-1,
        )
        mv = MedicalVolume(np.random.rand(*shape), np.eye(4), headers=headers)
        mv2 = MedicalVolume(mv.A, affine=mv.affine, headers=mv.headers())
        assert np.shares_memory(mv, mv)
        assert np.shares_memory(mv, mv2)

    def test_stack(self):
        shape = (10, 20, 30, 2)
        headers = np.stack(
            [
                ututils.build_dummy_headers(shape[2], {"EchoTime": 2}),
                ututils.build_dummy_headers(shape[2], {"EchoTime": 10}),
            ],
            axis=-1,
        )
        vol = np.ones(shape)
        mv_a = MedicalVolume(vol, np.eye(4), headers=headers)
        mv_b = MedicalVolume(vol * 2, np.eye(4), headers=headers)
        mv_c = MedicalVolume(vol * 3, np.eye(4), headers=headers)

        mv2 = np.stack([mv_a, mv_b, mv_c], axis=-1)
        assert mv2.shape == (10, 20, 30, 2, 3)
        assert mv2.headers() is not None
        assert np.all(mv2.volume == np.stack([mv_a.volume, mv_b.volume, mv_c.volume], axis=-1))
        mv2 = np.stack([mv_a, mv_b, mv_c], axis=-2)
        assert mv2.shape == (10, 20, 30, 3, 2)
        with self.assertRaises(ValueError):
            mv2 = np.stack([mv_a, mv_b, mv_c], axis=0)
        with self.assertRaises(TypeError):
            mv2 = np.stack([mv_a, mv_b, mv_c], axis=(-1,))

    def test_expand_dims(self):
        mv = self._build_volume()
        mv2 = np.expand_dims(mv, (-2, -3))
        assert mv2.shape == (10, 20, 30, 1, 1, 2)
        mv2 = np.expand_dims(mv, -1)
        assert mv2.shape == (10, 20, 30, 2, 1)
        with self.assertRaises(ValueError):
            mv2 = np.expand_dims(mv, 0)

    def test_squeeze(self):
        mv = self._build_volume()
        mv2 = mv[..., :1]
        assert mv2.shape == (10, 20, 30, 1)
        assert np.squeeze(mv2).shape == (10, 20, 30)
        assert np.squeeze(mv2, axis=-1).shape == (10, 20, 30)
        assert np.squeeze(mv2, axis=3).shape == (10, 20, 30)
        assert np.squeeze(mv2, axis=3).shape == (10, 20, 30)

        mv2 = mv[:1, :, :, :1]
        assert np.squeeze(mv2).shape == (1, 20, 30)
        with self.assertRaises(ValueError):
            np.squeeze(mv2, axis=0)

    def test_concatenate(self):
        shape = (10, 20, 30, 2)
        headers = np.stack(
            [
                ututils.build_dummy_headers(shape[2], {"EchoTime": 2}),
                ututils.build_dummy_headers(shape[2], {"EchoTime": 10}),
            ],
            axis=-1,
        )
        vol = np.ones(shape)
        mv_a = MedicalVolume(vol, np.eye(4), headers=headers)
        mv_b = MedicalVolume(vol * 2, np.eye(4), headers=headers)

        with self.assertRaises(ValueError):
            np.concatenate([mv_a, mv_b], axis=0)
        with self.assertRaises(TypeError):
            np.concatenate([mv_a, mv_b], axis="-1")

        mv2 = np.concatenate([mv_a, mv_b], axis=-1)
        assert np.all(mv2.volume == np.concatenate([mv_a.volume, mv_b.volume], axis=-1))
        assert mv2.headers().shape == (1, 1, 30, 4)

        affine = np.eye(4)
        affine[:3, 3] += [10, 0, 0]
        mv_d = MedicalVolume(vol * 2, affine, headers=headers)
        with self.assertWarns(UserWarning):
            mv2 = np.concatenate([mv_a, mv_d], axis=0)
        assert mv2.headers() is None
        assert np.all(mv2.volume == np.concatenate([mv_a.volume, mv_d.volume], axis=0))

        affine = np.eye(4)
        affine[:3, 3] += [0, 0, 30]
        mv_d = MedicalVolume(vol * 2, affine, headers=headers)
        mv2 = np.concatenate([mv_a, mv_d], axis=-2)
        assert np.all(mv2.volume == np.concatenate([mv_a.volume, mv_d.volume], axis=-2))
        assert mv2.headers().shape == (1, 1, 60, 2)

        affine = np.eye(4)
        affine[:3, 3] += [0, 0, 30]
        affine[:, 0] *= 0.5
        mv_d = MedicalVolume(vol * 2, affine, headers=headers)
        with self.assertRaises(ValueError):
            mv2 = np.concatenate([mv_a, mv_d], axis=-2)

        affine = np.eye(4)
        affine[:3, 3] += [0, 0, 30]
        mv_d = MedicalVolume(vol * 2, affine, headers=None)
        mv2 = np.concatenate([mv_a, mv_d], axis=-2)
        assert np.all(mv2.volume == np.concatenate([mv_a.volume, mv_d.volume], axis=-2))
        assert mv2.headers() is None

    def test_pad(self):
        shape = (10, 20, 30, 2)
        affine = np.eye(4)
        mv = self._build_volume(shape, affine=affine)

        # Pad all dims.
        pad_width = (1, 2, 3, 4)
        np_pad_width = ((1,), (2,), (3,), (4,))
        expected_origin = np.asarray([x - p for x, p in zip(mv.scanner_origin, pad_width)])
        expected_arr = np.pad(mv.A, np_pad_width)
        mv2 = np.pad(mv, pad_width)
        assert np.all(mv2.A == expected_arr)
        assert np.all(np.asarray(mv2.scanner_origin) == expected_origin)

        # Pad all dims by 1 on both sides.
        pad_width = 1
        np_pad_width = 1
        expected_origin = np.asarray([x - pad_width for x in mv.scanner_origin])
        expected_arr = np.pad(mv.A, np_pad_width)
        mv2 = np.pad(mv, pad_width)
        assert np.all(mv2.A == expected_arr)
        assert np.all(np.asarray(mv2.scanner_origin) == expected_origin)

        # Pad some dims.
        pad_width = (None, 0, 3, 4)
        np_pad_width = ((0,), (0,), (3,), (4,))
        expected_origin = np.asarray(
            [x - p if isinstance(p, int) else x for x, p in zip(mv.scanner_origin, pad_width)]
        )
        expected_arr = np.pad(mv.A, np_pad_width)
        mv2 = np.pad(mv, pad_width)
        assert np.all(mv2.A == expected_arr)
        assert np.all(np.asarray(mv2.scanner_origin) == expected_origin)

        pad_width = (3, 4)
        eff_pad_width = (0, 0, 3, 4)
        np_pad_width = ((0,), (0,), (3,), (4,))
        expected_origin = np.asarray(
            [x - p if isinstance(p, int) else x for x, p in zip(mv.scanner_origin, eff_pad_width)]
        )
        expected_arr = np.pad(mv.A, np_pad_width)
        mv2 = np.pad(mv, pad_width)
        assert np.all(mv2.A == expected_arr)
        assert np.all(np.asarray(mv2.scanner_origin) == expected_origin)

    def test_contiguous(self):
        arr = np.ones((2, 3, 4))

        mv = MedicalVolume(arr, np.eye(4))
        assert mv.is_contiguous()

        ornt = mv.orientation
        ornt = ornt[1:] + ornt[:1]
        mv = mv.reformat(ornt)
        assert not mv.is_contiguous()

        mv_contiguous = mv.contiguous()
        assert mv_contiguous.is_contiguous()
        assert mv_contiguous.A.data.contiguous
